"""Implementation of a StructureWrangler.

A StructureWrangler is used to generate and organize training data to fit a
cluster expansion using the terms defined in a ClusterSubspace. It takes care
of computing the training features (correlations) to construct a feature matrix
to be used along with a target property vector to obtain the coefficients for
a cluster expansion using some linear regression model.

Includes functions used to preprocess and check (wrangling) fitting data of
structures and properties.
"""

__author__ = "Luis Barroso-Luque"
__credits__ = "William Davidson Richard"

import warnings
from importlib import import_module
from itertools import combinations
from typing import Sequence

import numpy as np
from monty.json import MSONable, jsanitize
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core.structure import Structure
from pymatgen.entries.computed_entries import ComputedStructureEntry

from smol.exceptions import StructureMatchError


class StructureWrangler(MSONable):
    """Class to create fitting data to fit a cluster expansion.

    A StructureWrangler handles (wrangles) input data structures and properties
    to fit in a cluster expansion. This class holds a ClusterSubspace used to
    compute correlation vectors and produce feature/design matrices used to fit
    the final ClusterExpansion.

    This class is meant to take all input training data in the form of
    (structure, properties) where the properties represent the target
    material property for the given structure that will be used to train
    the cluster expansion, and returns the fitting data as a cluster
    correlation feature matrix (orbit basis function values).

    This class also has methods to check/prepare/filter the data. A metadata
    dictionary is used to keep track of applied filters, but users can also use
    it to save any other pertinent information that will be saved with using
    :code:`StructureWrangler.as_dict` for future reference. Other preprocessing
    and filtering methods are available in tools.py and select.py.
    """

    def __init__(self, cluster_subspace):
        """Initialize a StructureWrangler.

        Args:
            cluster_subspace (ClusterSubspace):
                a ClusterSubspace object that will be used to fit a
                ClusterExpansion with the provided data.
        """
        self._subspace = cluster_subspace
        self._entries = []
        self._ind_sets = {}  # data indices for test/training splits etc
        self._metadata = {"applied_filters": []}

    @property
    def cluster_subspace(self):
        """Get the underlying ClusterSubspace used to compute features."""
        return self._subspace

    @property
    def num_structures(self):
        """Get number of structures added (correctly matched to prim)."""
        return len(self._entries)

    @property
    def num_features(self):
        """Get number of features for each added structure."""
        return self.feature_matrix.shape[1]

    @property
    def available_properties(self):
        """Get list of properties that have been added."""
        return list(
            {p for entry in self._entries for p in entry.data["properties"].keys()}
        )

    @property
    def available_indices(self):
        """Get list of available data index sets."""
        return list(self._ind_sets.keys())

    @property
    def available_weights(self):
        """Get list of weights that have been added."""
        return list(
            {p for entry in self._entries for p in entry.data["weights"].keys()}
        )

    @property
    def structures(self):
        """Get list of included structures."""
        return [entry.structure for entry in self._entries]

    @property
    def refined_structures(self):
        """Get list of refined structures."""
        return [entry.data["refined_structure"] for entry in self._entries]

    @property
    def feature_matrix(self):
        """Get feature matrix.

        Rows are structures, and columns are correlation vectors.
        """
        return np.array([entry.data["correlations"] for entry in self._entries])

    @property
    def sizes(self):
        """Get sizes of each structure in terms of number of prims."""
        return np.array([entry.data["size"] for entry in self._entries])

    @property
    def occupancy_strings(self):
        """Get occupancy strings for each structure in the StructureWrangler."""
        occupancies = [
            self._subspace.occupancy_from_structure(
                entry.structure,
                entry.data["supercell_matrix"],
                entry.data["site_mapping"],
            )
            for entry in self._entries
        ]
        return occupancies

    @property
    def supercell_matrices(self):
        """Get list of supercell matrices relating each structure to prim."""
        return np.array([entry.data["supercell_matrix"] for entry in self._entries])

    @property
    def structure_site_mappings(self):
        """Get list of site mappings for each structure to prim."""
        return [entry.data["site_mapping"] for entry in self._entries]

    @property
    def entries(self):
        """Get a list of the entry dictionaries."""
        return self._entries

    def data_indices(self, key):
        """Get a specific data index set."""
        return self._ind_sets[key]

    @property
    def metadata(self):
        """Get dictionary to save applied filters, etc."""
        return self._metadata

    def get_feature_matrix_rank(self, rows=None, cols=None):
        """Get the rank of the feature matrix or a submatrix of it.

        Args:
            rows (list):
                indices of structures to include in feature matrix.
            cols (list):
                indices of features (correlations) to include in feature matrix

        Returns:
            int: the rank of the matrix
        """
        rows = rows if rows is not None else range(self.num_structures)
        cols = cols if cols is not None else range(self.num_features)
        return np.linalg.matrix_rank(self.feature_matrix[rows][:, cols])

    def get_feature_matrix_orbit_rank(self, orbit_id, rows=None):
        """Get the rank of an orbit submatrix of the feature matrix.

        Args:
            orbit_id (int):
                Orbit id to obtain sub feature matrix rank of.
            rows (list): optional
                list of row indices corresponding to structures to include.

        Returns:
            int: rank of orbit sub feature matrix
        """
        columns = [
            i
            for i, oid in enumerate(self._subspace.function_orbit_ids)
            if oid == orbit_id
        ]
        return self.get_feature_matrix_rank(rows=rows, cols=columns)

    def get_condition_number(self, rows=None, cols=None, norm_p=2):
        """Compute the condition number for the feature matrix or submatrix.

        The condition number is a measure of how sensitive the solution to
        the linear system is to perturbations in the sampled data. The larger
        the condition number the more ill-conditioned the linear problem is.

        Args:
            rows (list):
                indices of structures to include in feature matrix.
            cols (list):
                indices of features (correlations) to include in feature matrix
            norm_p : (optional)
                the type of norm to use when computing condition number.
                See the numpy docs for np.linalg.cond for options.

        Returns:
            float: matrix condition number
        """
        rows = rows if rows is not None else range(self.num_structures)
        cols = cols if cols is not None else range(self.num_features)
        return np.linalg.cond(self.feature_matrix[rows][:, cols], p=norm_p)

    def get_gram_matrix(self, rows=None, cols=None, normalize=True):
        r"""Compute the Gram matrix for the feature matrix or a submatrix.

        The Gram matrix, :math:`G = X^TX`, where each entry is
        :math:`G_{ij} = X_i \cdot X_j`. By default, G will have each column
        (feature vector) normalized. This makes it possible to compare Gram
        matrices for different feature matrix size or using different basis
        sets. This ensures every entry satisfies :math:`-1 \le G_{ij} \le 1`.

        Args:
            rows (list):
                indices of structures to include in feature matrix.
            cols (list):
                indices of features (correlations) to include in feature matrix
            normalize:
                if True (default), will normalize each feature vector in the
                feature matrix.
        Returns:
            ndarray: Gram matrix
        """
        rows = rows if rows is not None else range(self.num_structures)
        cols = cols if cols is not None else range(self.num_features)
        matrix = self.feature_matrix[rows][:, cols]
        if normalize:
            matrix /= np.sqrt(matrix.T.dot(matrix).diagonal())
        return matrix.T.dot(matrix)

    def get_duplicate_corr_indices(
        self, cutoffs=None, decimals=12, rm_external_terms=True
    ):
        """Find indices of rows with duplicate corr vectors in feature matrix.

        Args:
            cutoffs (dict): optional
                dictionary with cluster diameter cutoffs for correlation
                functions to consider in correlation vectors.
            decimals (int): optional
                number of decimals to round correlations in order to allow
                some numerical tolerance for finding duplicates. If None,
                no rounding will be done. Beware that using a ClusterSubspace
                with an orthogonal site basis will likely be off by some
                numerical tolerance so rounding is recommended.
            rm_external_terms (bool) : optional
                if True, will not consider external terms and only consider
                correlations proper when looking for duplicates.
        Returns:
            list: list containing lists of indices of rows in feature_matrix
            where duplicates occur
        """
        if len(self.feature_matrix) == 0:
            duplicate_inds = []
        else:
            num_ext = len(self.cluster_subspace.external_terms)
            if cutoffs is not None:
                inds = self._subspace.function_inds_from_cutoffs(cutoffs)
            else:
                inds = range(self.num_features - num_ext)
            if not rm_external_terms:
                inds.extend(range(-1, -num_ext - 1, -1))

            feature_matrix = (
                self.feature_matrix
                if decimals is None
                else np.around(
                    self.feature_matrix, decimals, self.feature_matrix.copy()
                )
            )
            _, inverse = np.unique(feature_matrix[:, inds], return_inverse=True, axis=0)
            duplicate_inds = [
                list(np.where(inverse == i)[0])
                for i in np.unique(inverse)
                if len(np.where(inverse == i)[0]) > 1
            ]
        return duplicate_inds

    def get_matching_corr_duplicate_indices(
        self, decimals=12, structure_matcher=None, **matcher_kwargs
    ):
        """Find indices of equivalent structures.

        Args:
            decimals (int): optional
                number of decimals to round correlations in order to allow
                some numerical tolerance for finding duplicates.
            structure_matcher (StructureMatcher): optional
                StructureMatcher object to use for matching structures.
            **matcher_kwargs:
                keyword arguments to use when initializing a StructureMatcher
                if not given

        Returns:
            list: list of lists of equivalent structures (that match) and have
            duplicate correlation vectors.
        """
        matcher = (
            structure_matcher
            if structure_matcher is not None
            else StructureMatcher(**matcher_kwargs)
        )
        duplicate_indices = self.get_duplicate_corr_indices(decimals)

        matching_inds = []
        for inds in duplicate_indices:
            # match all combinations of duplicates
            matches = [
                set(c)
                for c in combinations(inds, 2)
                if matcher.fit(
                    self.structures[c[0]], self.structures[c[1]], symmetric=True
                )
            ]
            overlaps = list(filter(lambda s: s[0] & s[1], combinations(matches, 2)))
            while overlaps:  # change to := walrus when commiting to 3.8 only
                all_overlaps = [o for overlap in overlaps for o in overlap]
                # keep only disjoint sets
                matches = [s for s in matches if s not in all_overlaps]
                # add union of overlapping sets
                for set1, set2 in overlaps:
                    if set1 | set2 not in matches:
                        matches.append(set1 | set2)
                overlaps = list(filter(lambda s: s[0] & s[1], combinations(matches, 2)))
            matching_inds += [list(sorted(m)) for m in matches]
        return matching_inds

    def get_constant_features(self):
        """Find indices of constant feature vectors (columns).

        A constant feature vector means the corresponding correlation function
        evaluates to the exact same value for all included structures, meaning
        it does not really help much when fitting. Many constant feature
        vectors may be a sign of insufficient sampling of configuration space.

        Excludes the empty cluster, which is by definition constant.

        Returns:
            ndarray: array of column indices.
        """
        arr = self.feature_matrix
        col_mask = np.all(arr == arr[0, :], axis=0)
        return np.where(col_mask == 1)[0][1:]

    def get_similarity_matrix(self, rows=None, cols=None, rtol=1e-5):
        """Get similarity matrix of correlation vectors.

        Generate a matrix to compare the similarity of correlation feature
        vectors (columns) in the feature matrix. Matrix element a(i,j)
        represents the fraction of equivalent corresponding values in feature
        vectors i and j.
        This construction is analogous to the Gram matrix, but instead of an
        inner product, it counts the number of identical corresponding elements
        in feature vectors i and j.

        Args:
            rows (list):
                indices of structures to include in feature matrix.
            cols (list):
                indices of features (correlations) to include in feature matrix
            rtol (float):
                relative tolerance for comparing feature matrix column values

        Returns:
            ndarray:
                (n x n) similarity matrix
        """
        rows = rows if rows is not None else range(self.num_structures)
        cols = cols if cols is not None else range(self.num_features)

        matrix = self.feature_matrix[rows][:, cols]
        num_structs = len(rows)
        num_corrs = len(cols)
        sim_matrix = np.identity(num_corrs)
        for i in range(num_corrs):
            for j in range(i + 1, num_corrs):
                num_identical = np.sum(
                    np.isclose(matrix[:, i], matrix[:, j], rtol=rtol)
                )
                sim_matrix[i, j] = num_identical / num_structs
                sim_matrix[j, i] = num_identical / num_structs

        return sim_matrix

    def get_property_vector(self, key, normalize=True):
        """Get the property target vector.

        The property target vector to be used to fit the corresponding
        correlation feature matrix to obtain coefficients for a cluster
        expansion. It should always be properly/consistently normalized when
        used for a fit.

        Args:
            key (str):
                name of the property
            normalize (bool): optional
                if True, normalizes by prim size. If the property sought is not
                already normalized, you need to normalize before fitting a CE.
        """
        if key in self.available_properties:
            properties = np.array(
                [entry.data["properties"][key] for entry in self._entries]
            )
        else:
            try:
                properties = np.array([getattr(entry, key) for entry in self._entries])
            except AttributeError as error:
                raise ValueError(
                    f"{key} is not a valid property. The property must be a valid "
                    f"ComputedStructureEntry property or one of "
                    f"{self.available_properties}"
                ) from error

        if normalize:
            properties /= self.sizes

        return properties

    def add_data_indices(self, key, indices):
        """Add a set of data indices.

        For example, use this for saving test/training splits or separating
        duplicates.
        """
        if not isinstance(indices, (Sequence, np.ndarray)):
            raise TypeError("indices must be Sequence like or an ndarray.")

        if any(i not in range(self.num_structures) for i in indices):
            raise ValueError("One or more indices are out of range.")
        self._ind_sets[key] = list(indices)

    def get_weights(self, key):
        """Get the weights specified by the given key.

        Args:
            key (str):
                name of corresponding weights
        """
        return np.array([entry.data["weights"][key] for entry in self._entries])

    def add_entry(
        self,
        entry,
        properties=None,
        weights=None,
        supercell_matrix=None,
        site_mapping=None,
        verbose=True,
        raise_failed=False,
    ):
        """Add a structure and measured property to the StructureWrangler.

        The energy and properties need to be extensive (i.e. not normalized per atom
        or unit cell, directly from DFT).

        An attempt to compute the correlation vector is made and if successful the
        structure is succesfully added. Otherwise the structure is ignored.
        Usually failures are caused by the StructureMatcher in the given
        ClusterSubspace failing to map structures to the primitive structure.

        Args:
            entry (ComputedStructureEntry):
                A ComputedStructureEntry with a training structure, energy and
                properties
            properties (dict):
                Dictionary with a key describing the property and the target
                value for the corresponding structure. For example if only a
                single property {'energy': value} but can also add more than
                one, i.e. {'total_energy': value1, 'formation_energy': value2}.
                You are free to make up the keys for each property but make
                sure you are consistent for all structures that you add.
            weights (dict):
                the weight given to the structure when doing the fit. The key
                must match at least one of the given properties.
            supercell_matrix (ndarray): optional
                if the corresponding structure has already been matched to the
                ClusterSubspace prim structure, passing the supercell_matrix
                will use that instead of trying to re-match. If using this,
                the user is responsible for having the correct supercell_matrix.
                Here you are the cause of your own bugs.
            site_mapping (list): optional
                site mapping as obtained by StructureMatcher.get_mapping
                such that the elements of site_mapping represent the indices
                of the matching sites to the prim structure. If you pass this
                option, you are fully responsible that the mappings are correct!
            verbose (bool): optional
                if True, will raise warning regarding  structures that fail in
                StructureMatcher, and structures that have duplicate corr vectors.
            raise_failed (bool): optional
                if True, will raise the thrown error when adding a structure
                that  fails. This can be helpful to keep a list of structures that
                fail for further inspection.
        """
        processed_entry = self.process_entry(
            entry,
            properties,
            weights,
            supercell_matrix,
            site_mapping,
            verbose,
            raise_failed,
        )
        if processed_entry is not None:
            self._entries.append(processed_entry)
        if verbose:
            self._corr_duplicate_warning(self.num_structures - 1)

    def append_entries(self, entries):
        """Append a list of entrys.

        Each entry must have all necessary fields. A entry can be
        obtained using the process_structure method.

        Args:
            entries (list of ComputedStructureEntry):
                list of entrys with all necessary information
        """
        required_keys = [
            "refined_structure",
            "supercell_matrix",
            "site_mapping",
            "correlations",
            "size",
        ]

        for i, entry in enumerate(entries):
            if not all(key in entry.data.keys() for key in required_keys):
                raise ValueError(
                    f"entry {i} is missing required keys. Make sure"
                    " they were obtained with the process_structure method."
                )
            if len(self._entries) > 0:
                if set(entry.data["properties"].keys()) != set(
                    self.available_properties
                ):
                    raise ValueError(
                        f"The properties in the entry being added do not match all "
                        f"the properties that have already been added: "
                        f"{self.available_properties}.\n Additional entrys must include "
                        f"the same properties included."
                    )

                if set(entry.data["weights"].keys()) != set(self.available_weights):
                    raise ValueError(
                        f"The properties in the entry being added do not match all "
                        f"the weights that have already been added: "
                        f"{self.available_weights}.\n Additional entrys must include the"
                        f" same weights included."
                    )
            self._entries.append(entry)
            self._corr_duplicate_warning(self.num_structures - 1)

    def add_weights(self, key, weights):
        """Add weights to structures already in the StructureWrangler.

        The length of the given weights must match the number of structures
        contained, and should be in the same order.

        Args:
            key (str):
                name describing weights
            weights (ndarray):
                array with the weight for each structure
        """
        if self.num_structures != len(weights):
            raise AttributeError(
                "Length of weights must match number of structures "
                f"{len(weights)} != {self.num_structures}."
            )
        for weight, entry in zip(weights, self._entries):
            entry.data["weights"][key] = weight

    def add_properties(self, key, property_vector):
        """Add another property vector to structures already in the StructureWrangler.

        The length of the property vector must match the number of structures
        contained, and should be in the same order so that the property
        corresponds to the correct structure.

        Args:
            key (str):
                name of property
            property_vector (ndarray):
                array with the property for each structure
        """
        if self.num_structures != len(property_vector):
            raise AttributeError(
                "Length of property_vector must match number of structures"
                f" {len(property_vector)} != {self.num_structures}."
            )

        for prop, entry in zip(property_vector, self._entries):
            entry.data["properties"][key] = prop

    def remove_properties(self, *property_keys):
        """Remove properties with given keys.

        Args:
            *property_keys (str):
                names of properties to remove
        """
        for key in property_keys:
            try:
                for entry in self._entries:
                    del entry.data["properties"][key]
            except KeyError:
                warnings.warn(f"Propertiy {key} does not exist.", RuntimeWarning)

    def remove_entry(self, entry):
        """Remove a given structure and associated data."""
        try:
            index = self._entries.index(entry)
            del self._entries[index]
        except ValueError:
            warnings.warn(
                f"Entry {entry} was not found. Nothing has been removed.",
                RuntimeWarning,
            )

    def change_subspace(self, cluster_subspace):
        """Change the underlying ClusterSubspace.

        Will swap out the ClusterSubspace and update features accordingly.
        This is a faster operation than creating a new one. Can also be useful
        to create a copy and then change the subspace.

        Args:
            cluster_subspace:
                New ClusterSubspace to be used for determining features.
        """
        self._subspace = cluster_subspace
        self.update_features()

    def update_features(self):
        """Update the features/feature matrix for the data held.

        This is useful when something is changed in the ClusterSubspace after
        creating the StructureWrangler, for example when adding an Ewald term
        after the StructureWrangler has already been created. This will prevent
        having to re-match structures and such.
        """
        for entry in self._entries:
            structure = entry.structure
            scmatrix = entry.data["supercell_matrix"]
            mapping = entry.data["site_mapping"]
            entry.data["correlations"] = self._subspace.corr_from_structure(
                structure, scmatrix=scmatrix, site_mapping=mapping
            )

    def remove_all_data(self):
        """Remove all data from the StructureWrangler."""
        self._entries = []

    def process_entry(
        self,
        entry,
        properties=None,
        weights=None,
        supercell_matrix=None,
        site_mapping=None,
        verbose=False,
        raise_failed=False,
    ):
        """Process a ComputedStructureEntry to be added to StructureWrangler.

        Checks if the structure for this entry can be matched to the
        ClusterSubspace prim structure to obtain its supercell matrix,
        correlation, and refined structure. If so, the entry will be updated by adding
        these to its data dictionary.

        Args:
            entry (ComputedStructureEntry):
                A ComputedStructureEntry corresponding to a training strucutre and
                properties
            properties (dict): optional
                A dictionary with a keys describing the property and the target
                value for the corresponding structure. Energy and corrected energy
                should already be in the ComputedStructureEntry so there is no need
                to pass it here.
                You are free to make up the keys for each property but make
                sure you are consistent for all structures that you add.
            weights (dict): optional
                The weight given to the structure when doing the fit. The key
                must match at least one of the given properties.
            supercell_matrix (ndarray): optional
                if the corresponding structure has already been matched to the
                ClusterSubspace prim structure, passing the supercell_matrix
                will use that instead of trying to re-match. If using this
                the user is responsible to have the correct supercell_matrix.
                Here you are the cause of your own bugs.
            site_mapping (list): optional
                site mapping as obtained by
                :code:`StructureMatcher.get_mapping`
                such that the elements of site_mapping represent the indices
                of the matching sites to the prim structure. If you pass this
                option, you are fully responsible that the mappings are correct!
            verbose (bool):
                if True, will raise warning for structures that fail in
                StructureMatcher, and structures that have duplicate corr vectors.
            raise_failed (bool): optional
                if True, will raise the thrown error when adding a structure
                that fails. This can be helpful to keep a list of structures that
                fail for further inspection.

        Returns:
            ComputedStructureEntry: entry with CE pertinent properties
        """
        if len(self._entries) > 0:
            if properties is not None and len(self.available_properties) > 0:
                if set(properties.keys()) != set(self.available_properties):
                    print(set(properties.keys()))
                    raise ValueError(
                        f"The properties entry being added do not match all "
                        f"the properties that have already been added: "
                        f"{self.available_properties}.\n Additional entries must "
                        f"include the same properties included."
                    )

            if weights is not None and len(self.available_properties) > 0:
                if set(weights.keys()) != set(self.available_weights):
                    raise ValueError(
                        f"The properties in the entry being added do not match all the"
                        f"weights that have already been added: "
                        f"{self.available_weights}.\n Additional entries must include "
                        f"the same weights included."
                    )

        properties = {} if properties is None else properties
        weights = {} if weights is None else weights

        try:
            if supercell_matrix is None:
                supercell_matrix = self._subspace.scmatrix_from_structure(
                    entry.structure
                )

            size = self._subspace.num_prims_from_matrix(supercell_matrix)
            if site_mapping is None:
                supercell = self._subspace.structure.copy()
                supercell.make_supercell(supercell_matrix)
                site_mapping = self._subspace.structure_site_mapping(
                    supercell, entry.structure
                )

            fm_row = self._subspace.corr_from_structure(
                entry.structure, scmatrix=supercell_matrix, site_mapping=site_mapping
            )
            refined_struct = self._subspace.refine_structure(
                entry.structure, supercell_matrix
            )

        except StructureMatchError as s_match_error:
            if verbose:
                warnings.warn(
                    f"Unable to match {entry.composition} with "
                    f"properties {properties} to supercell_structure. "
                    f"Throwing out.\n Error Message: {str(s_match_error)}",
                    UserWarning,
                )
            if raise_failed:
                raise s_match_error
            return None

        entry.data["refined_structure"] = refined_struct
        entry.data["supercell_matrix"] = supercell_matrix
        entry.data["site_mapping"] = site_mapping
        entry.data["correlations"] = fm_row
        entry.data["size"] = size
        property_dict = entry.data.get("properties")
        if property_dict is not None:
            property_dict.update(properties)
        else:
            entry.data["properties"] = properties

        weight_dict = entry.data.get("weights")
        if weight_dict is not None:
            weight_dict.update(weights)
        else:
            entry.data["weights"] = weights

        return entry

    def _corr_duplicate_warning(self, index):
        """Warn if corr vector of entry with given index is duplicated."""
        for duplicate_inds in self.get_duplicate_corr_indices():
            if index in duplicate_inds:
                duplicates = "".join(
                    f"Index {i} - {self._entries[i].structure.composition} "
                    f"energy={self._entries[i].energy}\n"
                    for i in duplicate_inds
                )
                warnings.warn(
                    "The following structures have duplicated correlation "
                    f"vectors:\n {duplicates} Consider adding more terms to "
                    "the clustersubspace or filtering duplicates.",
                    UserWarning,
                )

    @classmethod
    def from_dict(cls, d, energy_key=None):
        """Create Structure Wrangler from an MSONable dict.

        Args:
            d (dict):
                MSON dict of StructureWrangler
            energy_key (str): optional
                energy property key, for legacy files

        Returns:
            StructureWrangler
        """
        module = import_module(d["_subspace"]["@module"])
        subspace_cls = getattr(module, d["_subspace"]["@class"])
        wrangler = cls(cluster_subspace=subspace_cls.from_dict(d["_subspace"]))

        # check for legacy files
        if "_items" in d.keys():
            warnings.warn(
                "This StructureWrangler was created with a previous version of smol.\n"
                "Please re-save it to prevent this warning.",
                FutureWarning,
            )
            if energy_key is None:
                raise RuntimeWarning(
                    "Please load using from_dict method and provide the key used for "
                    "the energy property to correctly load this StructureWrangler."
                )
            entries = []
            for item in d["_items"]:
                energy = item["properties"].pop(energy_key)
                data = {
                    "properties": item["properties"],
                    "weights": item["weights"],
                    "supercell_matrix": np.array(item["scmatrix"]),
                    "correlations": np.array(item["features"]),
                    "size": item["size"],
                    "site_mapping": item["mapping"],
                    "refined_structure": Structure.from_dict(item["ref_structure"]),
                }
                entries.append(
                    ComputedStructureEntry(
                        Structure.from_dict(item["structure"]), energy, data=data
                    )
                )
        else:
            entries = [
                ComputedStructureEntry.from_dict(entry_d) for entry_d in d["_entries"]
            ]

        wrangler._entries = entries
        wrangler._metadata = d["metadata"]
        wrangler._ind_sets = d.get("_ind_sets") or {}
        return wrangler

    def as_dict(self):
        """Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        wrangler_dict = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "_subspace": self._subspace.as_dict(),
            "_entries": [entry.as_dict() for entry in self._entries],
            "_ind_sets": jsanitize(self._ind_sets),  # jic for np.int's
            "metadata": self.metadata,
        }
        return wrangler_dict



"""Implementation of Semi-Grand Canonical Ensemble classes.

These are used to run Monte Carlo sampling for systems with
a fixed number of sites but variable concentration of species.
"""

__author__ = "Luis Barroso-Luque"


from collections import Counter

import numpy as np
from monty.json import MSONable
from pymatgen.core import DummySpecies, Element, Species

from smol.cofe.space.domain import Vacancy, get_species
from smol.moca.processor.base import Processor
from smol.moca.sublattice import Sublattice

from .base import Ensemble


class SemiGrandEnsemble(Ensemble, MSONable):
    """Relative chemical potential-based SemiGrand Ensemble.

    A Semi-Grand Canonical Ensemble for Monte Carlo simulations where species'
    relative chemical potentials are predefined. Note that in the SGC Ensemble
    implemented here, only the differences in chemical potentials with
    respect to a reference species on each sublattice are fixed, and not the
    absolute values. To obtain the absolute values you must calculate the
    reference chemical potential and then subtract it from the given values.

    Attributes:
        thermo_boundaries (dict):
            dict of chemical potentials.
    """

    valid_mcmc_steps = ("flip",)

    def __init__(self, processor, chemical_potentials, sublattices=None):
        """Initialize MuSemiGrandEnsemble.

        Args:
            processor (Processor):
                a processor that can compute the change in property given
                a set of flips.
            chemical_potentials (dict):
                Dictionary with species and chemical potentials.
            sublattices (list of Sublattice): optional
                list of Sublattice objects representing sites in the processor
                supercell with same site spaces.
        """
        super().__init__(processor, sublattices=sublattices)
        self._params = np.append(self.processor.coefs, -1.0)
        # check that species are valid
        chemical_potentials = {
            get_species(k): v for k, v in chemical_potentials.items()
        }
        # Excessive species not appeared on active sub-lattices
        # will be dropped.
        for spec in self.species:
            if spec not in chemical_potentials.keys():
                raise ValueError(
                    f"Species {spec} was not assigned a chemical "
                    " potential, a value must be provided."
                )

        # preallocate this for slight speed improvements
        self._dfeatures = np.empty(len(processor.coefs) + 1)
        self._features = np.empty(len(processor.coefs) + 1)

        self._mus = {k: v for k, v in chemical_potentials.items() if k in self.species}
        self._mu_table = self._build_mu_table(self._mus)
        self.thermo_boundaries = {
            "chemical-potentials": {str(k): v for k, v in self._mus.items()}
        }

    @property
    def natural_parameters(self):
        """Get the vector of natural parameters.

        For SGC an extra -1 is added for the chemical part of the Legendre
        transform.
        """
        return self._params

    @property
    def chemical_potentials(self):
        """Get the chemical potentials for species in the system."""
        return self._mus

    @chemical_potentials.setter
    def chemical_potentials(self, value):
        """Set the chemical potentials and update table."""
        for spec, count in Counter(map(get_species, value.keys())).items():
            if count > 1:
                raise ValueError(
                    f"{count} values of the chemical potential for the same "
                    f"species {spec} were provided.\n Make sure the dictionary "
                    "you are using has only string keys or only Species "
                    "objects as keys."
                )
        value = {get_species(k): v for k, v in value.items() if k in self.species}
        if set(value.keys()) != set(self.species):
            raise ValueError(
                "Chemical potentials given are missing species. "
                "Values must be given for each of the following:"
                f" {self.species}"
            )
        self._mus = value
        self._mu_table = self._build_mu_table(value)
        self.thermo_boundaries = {
            "chemical-potentials": {str(k): v for k, v in self._mus.items()}
        }

    def compute_feature_vector(self, occupancy):
        """Compute the relevant feature vector for a given occupancy.

        In the semigrand case it is the feature vector and the chemical work
        term.

        Args:
            occupancy (ndarray):
                encoded occupancy string

        Returns:
            ndarray: feature vector
        """
        self._features[:-1] = self.processor.compute_feature_vector(occupancy)
        self._features[-1] = self.compute_chemical_work(occupancy)
        return self._features

    def compute_feature_vector_change(self, occupancy, step):
        """Return the change in the feature vector from a given step.

        Args:
            occupancy (ndarray):
                encoded occupancy string.
            step (list of tuple):
                a sequence of flips given by MCUsher.propose_step

        Returns:
            ndarray: difference in feature vector
        """
        self._dfeatures[:-1] = self.processor.compute_feature_vector_change(
            occupancy, step
        )
        self._dfeatures[-1] = sum(
            self._mu_table[f[0]][f[1]] - self._mu_table[f[0]][occupancy[f[0]]]
            for f in step
        )
        return self._dfeatures

    def compute_chemical_work(self, occupancy):
        """Compute sum of mu * N for given occupancy."""
        return sum(
            self._mu_table[site][species] for site, species in enumerate(occupancy)
        )

    def split_sublattice_by_species(self, sublattice_id, occu, species_in_partitions):
        """Split a sub-lattice in system by its occupied species.

        An example use case might be simulating topotactic Li extraction
        and insertion, where we want to consider Li/Vac, TM and O as
        different sub-lattices that can not be mixed by swapping.

        In the grand canonical ensemble, the mu table will also be updated
        after split.

        Args:
            sublattice_id (int):
                The index of sub-lattice to split in self.sublattices.
            occu (np.ndarray[int]):
                An occupancy array to reference with.
            species_in_partitions (List[List[int|Species|Vacancy|Element|str]]):
                Each sub-list contains a few species or encodings of species in
                the site space to be grouped as a new sub-lattice, namely,
                sites with occu[sites] == specie in the sub-list, will be
                used to initialize a new sub-lattice.
                Sub-lists will be pre-sorted to ascending order.
        """
        super().split_sublattice_by_species(sublattice_id, occu, species_in_partitions)
        # Species in active sub-lattices may change after split.
        # Need to reset and rebuild mu table.
        new_chemical_potentials = {spec: self._mus[spec] for spec in self.species}
        self.chemical_potentials = new_chemical_potentials

    def _build_mu_table(self, chemical_potentials):
        """Build an array for chemical potentials for all sites in system.

        Rows represent sites and columns species. This allows quick evaluation
        of chemical potential changes from flips. Not that the total number
        of columns will be the number of species in the largest site space. For
        smaller site spaces the values at those rows are meaningless and will
        be given values of 0. Also rows representing sites with not partial
        occupancy will have all 0 values and should never be used.
        """
        # Mu table should be built with ensemble, rather than processor data.
        # Otherwise you may get wrong species encoding if the sub-lattices are
        # split.
        num_cols = max(max(sl.encoding) for sl in self.sublattices) + 1
        # Sublattice can only be initialized as default, or splitted from default.
        table = np.zeros((self.num_sites, num_cols))
        for sublatt in self.active_sublattices:
            ordered_pots = [chemical_potentials[sp] for sp in sublatt.site_space]
            table[sublatt.sites[:, None], sublatt.encoding] = ordered_pots
        return table

    def as_dict(self):
        """Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        sgce_d = super().as_dict()
        sgce_d["chemical_potentials"] = tuple(
            (s.as_dict(), c) for s, c in self.chemical_potentials.items()
        )
        return sgce_d

    @classmethod
    def from_dict(cls, d):
        """Instantiate a SemiGrandEnsemble from dict representation.

        Returns:
            CanonicalEnsemble
        """
        chemical_potentials = {}
        for spec, chem_pot in d["chemical_potentials"]:
            if "oxidation_state" in spec and Element.is_valid_symbol(spec["element"]):
                spec = Species.from_dict(spec)
            elif "oxidation_state" in spec:
                if spec["@class"] == "Vacancy":
                    spec = Vacancy.from_dict(spec)
                else:
                    spec = DummySpecies.from_dict(spec)
            else:
                spec = Element(spec["element"])
            chemical_potentials[spec] = chem_pot

        sublatts = d.get("sublattices")  # keep backwards compatibility
        if sublatts is not None:
            sublatts = [Sublattice.from_dict(sl_d) for sl_d in sublatts]

        return cls(
            Processor.from_dict(d["processor"]),
            chemical_potentials=chemical_potentials,
            sublattices=sublatts,
        )



"""Implementation of Composite processor class for a fixed size supercell.

If you have created a ClusterExpansion with additional terms, and don't want
the headache of manually and correctly setting up the corresponding
CompositeProcessor, simply consider using the convenience class constructor
method in the ensemble classes: Ensemble.from_clusterexpansion, that will take
care of all these hiccups for you.
"""

__author__ = "Luis Barroso-Luque"

import numpy as np

from smol.cofe.space.clusterspace import ClusterSubspace
from smol.moca.processor.base import Processor


class CompositeProcessor(Processor):
    """CompositeProcessor class used for mixed models.

    A Composite processor is merely a container for many different processors
    that acts as an interface such that it can be used in the same way as an
    individual processor. This can be used to mix models in any way that your
    heart desires.

    The most common use case is a CompositeProcessor of a
    ClusterExpansionProcessor and an EwaldProcessor for use in ionic materials
    when the cluster expansion contains an Ewald term.

    You can add any of the other processor class implemented to build a
    composite processor.

    It is recommended to use the :code:`from_cluster_expansion` to create an
    ensemble and the underlying processor automatically created rather than
    directly creating a processor. This will take care of creating the correct
    :class:`CompositeProcessor` or :class:`ClusterExpansionProcessor` for you.
    """

    def __init__(self, cluster_subspace, supercell_matrix):
        """Initialize a CompositeProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                a cluster subspace
            supercell_matrix (ndarray):
                an array representing the supercell matrix with respect to the
                cluster expansion prim structure.
        """
        super().__init__(cluster_subspace, supercell_matrix, None)
        self._processors = []
        self.coefs = np.empty(0)

    @property
    def processors(self):
        """Return the list of processors in CompositeProcessor."""
        return self._processors

    def add_processor(self, processor):
        """Add a processor to composite.

        Args:
            processor (Processor):
                processor to add. All processors must have the same
                ClusterSubspace and supercell matrix.
        """
        if isinstance(processor, CompositeProcessor):
            raise AttributeError(
                "A CompositeProcessor can not be added into "
                "another CompositeProcessor"
            )
        if self.cluster_subspace != processor.cluster_subspace:
            raise ValueError(
                "The cluster subspace of the processor to be"
                " added does not match the one of this "
                "CompositeProcessor."
            )
        if not np.array_equal(self._scmatrix, processor.supercell_matrix):
            raise ValueError(
                "The supercell matrix of the processor to be"
                " added does not match the one of this "
                "CompositeProcessor."
            )
        self._processors.append(processor)
        self.coefs = np.append(self.coefs, processor.coefs)

    def compute_property(self, occupancy):
        """Compute the value of the property for the given occupancy array.

        Args:
            occupancy (ndarray):
                encoded occupancy array
        Returns:
            float: predicted property
        """
        return sum(pr.compute_property(occupancy) for pr in self._processors)

    def compute_property_change(self, occupancy, flips):
        """Compute the change in property from a set of flips.

        Args:
            occupancy (ndarray):
                encoded occupancy array
            flips (list):
                list of tuples for (index of site, specie code to set)

        Returns:
            float:  property difference between inital and final states
        """
        return sum(
            pr.compute_property_change(occupancy, flips) for pr in self._processors
        )

    def compute_feature_vector(self, occupancy):
        """Compute the feature vector for a given occupancy array.

        Each entry in the correlation vector corresponds to a particular
        symmetrically distinct bit ordering.

        Args:
            occupancy (ndarray):
                encoded occupation array

        Returns:
            array: correlation vector
        """
        features = [
            np.array(pr.compute_feature_vector(occupancy)) for pr in self._processors
        ]
        # TODO you may be able to cut some speed by pre-allocating this
        return np.append(features[0], features[1:])

    def compute_feature_vector_change(self, occupancy, flips):
        """
        Compute the change in the feature vector from a list of flips.

        Args:
            occupancy (ndarray):
                encoded occupancy array
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a
                single flip where the first element is the index of the site
                in the occupancy array and the second element is the index
                for the new species to place at that site.

        Returns:
            array: change in feature vector
        """
        updates = [
            np.array(pr.compute_feature_vector_change(occupancy, flips))
            for pr in self._processors
        ]
        # TODO you may be able to cut some speed by pre-allocating this
        return np.append(updates[0], updates[1:])

    def as_dict(self) -> dict:
        """
        Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        proc_d = super().as_dict()
        proc_d["processors"] = [pr.as_dict() for pr in self._processors]
        return proc_d

    @classmethod
    def from_dict(cls, d):
        """Create a CompositeProcessor from serialized MSONable dict."""
        # pylint: disable=duplicate-code
        proc = cls(
            ClusterSubspace.from_dict(d["cluster_subspace"]),
            np.array(d["supercell_matrix"]),
        )
        for prd in d["processors"]:
            proc.add_processor(Processor.from_dict(prd))
        return proc



import warnings
from itertools import combinations

import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.analysis.structure_matcher import (
    OrderDisorderElementComparator,
    StructureMatcher,
)
from pymatgen.core import Species, Structure
from pymatgen.util.coord import is_coord_subset_pbc

from smol.cofe import ClusterSubspace, PottsSubspace
from smol.cofe.space.clusterspace import get_complete_mapping, invert_mapping
from smol.cofe.space.constants import SITE_TOL
from smol.cofe.space.domain import Vacancy, get_allowed_species
from smol.correlations import corr_from_occupancy
from smol.exceptions import StructureMatchError
from tests.utils import assert_msonable, gen_random_structure


def test_from_cutoffs(structure):
    cutoffs = {2: 5, 3: 4, 4: 4}
    for increment in np.arange(0, 3, 1):
        cutoffs.update(
            {k: v + increment / (n + 1) for n, (k, v) in enumerate(cutoffs.items())}
        )
        subspace = ClusterSubspace.from_cutoffs(structure, cutoffs)
        tight_subspace = ClusterSubspace.from_cutoffs(structure, subspace.cutoffs)
        assert len(subspace) == len(tight_subspace)
        npt.assert_allclose(
            np.array(list(subspace.cutoffs.values())),
            np.array(list(tight_subspace.cutoffs.values())),
        )


def test_orbits(cluster_subspace):
    # test that all orbits generated are unique
    assert len(cluster_subspace) == cluster_subspace.num_corr_functions
    assert len(cluster_subspace.orbits) + 1 == cluster_subspace.num_orbits

    for o1, o2 in combinations(cluster_subspace.orbits, 2):
        assert o1 != o2


def test_cutoffs(cluster_subspace, cluster_cutoffs):
    for s, c in cluster_subspace.cutoffs.items():
        assert cluster_cutoffs[s] >= c


def test_orbits_from_cutoffs(cluster_subspace, cluster_cutoffs):
    # Get all of them
    max_cutoff = max(cluster_cutoffs.values())
    assert all(
        o1 == o2
        for o1, o2 in zip(
            cluster_subspace.orbits, cluster_subspace.orbits_from_cutoffs(max_cutoff)
        )
    )
    for upper, lower in ((5, 0), (6, 3), (5, 2)):
        orbs = cluster_subspace.orbits_from_cutoffs(upper, lower)
        assert len(orbs) < len(cluster_subspace.orbits)
        assert all(lower <= o.base_cluster.diameter <= upper for o in orbs)

    # Test with dict
    upper = {2: 4.5, 3: 3.5}
    orbs = cluster_subspace.orbits_from_cutoffs(upper)
    assert len(orbs) < len(cluster_subspace.orbits)
    assert all(
        o.base_cluster.diameter <= upper[2] for o in orbs if len(o.base_cluster) == 2
    )
    assert all(
        o.base_cluster.diameter <= upper[3] for o in orbs if len(o.base_cluster) == 3
    )

    # Test for only pairs
    upper = {2: 4.5}
    orbs = cluster_subspace.orbits_from_cutoffs(upper)
    assert len(orbs) < len(cluster_subspace.orbits)
    assert all(
        o.base_cluster.diameter <= upper[2] for o in orbs if len(o.base_cluster) == 2
    )
    assert all(len(o.base_cluster) == 2 for o in orbs)

    # bad cuttoffs
    assert len(cluster_subspace.orbits_from_cutoffs(2, 4)) == 0


def test_functions_inds_by_size(cluster_subspace):
    indices = cluster_subspace.function_inds_by_size
    # check that all orbit functions are in there...
    assert sum(len(i) for i in indices.values()) == len(cluster_subspace) - 1
    fun_orb_ids = cluster_subspace.function_orbit_ids
    # Now check sizes are correct.
    for s, inds in indices.items():
        assert all(
            s == len(cluster_subspace.orbits[fun_orb_ids[i] - 1].base_cluster)
            for i in inds
        )


def test_functions_inds_by_cutoffs(cluster_subspace):
    indices = cluster_subspace.function_inds_from_cutoffs(6)
    # check that all of them are in there.
    assert len(indices) == len(cluster_subspace) - 1
    fun_orb_ids = cluster_subspace.function_orbit_ids
    for upper, lower in ((4, 0), (5, 3), (3, 1)):
        indices = cluster_subspace.function_inds_from_cutoffs(upper, lower)
        assert len(indices) < len(cluster_subspace)
        assert all(
            lower
            <= cluster_subspace.orbits[fun_orb_ids[i] - 1].base_cluster.diameter
            <= upper
            for i in indices
        )


@pytest.mark.parametrize("orthonormal", [(True, False)])
def test_site_bases(cluster_subspace, basis_name, orthonormal):
    subspace = cluster_subspace.copy()  # copy it to keep original state
    subspace.change_site_bases(basis_name, orthonormal=orthonormal)
    if orthonormal:
        assert subspace.basis_orthogonal
        assert subspace.basis_orthonormal

    structure = gen_random_structure(subspace.structure)
    if cluster_subspace.basis_type == subspace.basis_type and not orthonormal:
        npt.assert_array_almost_equal(
            subspace.corr_from_structure(structure),
            cluster_subspace.corr_from_structure(structure),
        )
    else:
        assert not np.allclose(
            subspace.corr_from_structure(structure),
            cluster_subspace.corr_from_structure(structure),
        )


# TODO These can probably be improved to check odd and specific cases we want
#  to watch out for
def test_supercell_matrix_from_structure(cluster_subspace, rng):
    # Simple scaling
    supercell = cluster_subspace.structure.copy()
    supercell.make_supercell(2)
    sc_matrix = cluster_subspace.scmatrix_from_structure(supercell)
    assert np.linalg.det(sc_matrix) == pytest.approx(8)

    # A more complex supercell_structure
    m = np.array([[0, 5, 3], [-2, 0, 2], [-2, 4, 3]])
    supercell = cluster_subspace.structure.copy()
    supercell.make_supercell(m)

    sc_matrix = cluster_subspace.scmatrix_from_structure(supercell)
    assert np.linalg.det(sc_matrix) == pytest.approx(abs(np.linalg.det(m)))

    # Test a slightly distorted structure
    supercell = cluster_subspace.structure.copy()
    # up to 2% strain
    supercell.apply_strain(rng.uniform(-0.02, 0.02, size=3))
    supercell.make_supercell(2)
    sc_matrix = cluster_subspace.scmatrix_from_structure(supercell)
    assert np.linalg.det(sc_matrix) == pytest.approx(8)


def test_refine_structure(cluster_subspace, rng):
    supercell = cluster_subspace.structure.copy()
    supercell.make_supercell(3)
    structure = gen_random_structure(cluster_subspace.structure, size=3)
    structure.apply_strain(rng.uniform(-0.01, 0.01, size=3))

    #  TODO this sometimes fails because sc_matrix is not found!!!
    try:
        refined_structure = cluster_subspace.refine_structure(structure)
    except StructureMatchError:
        warnings.warn(
            "StructureMatchError was raised when testing refined structures"
            f" for testing structure {cluster_subspace.structure}",
            RuntimeWarning,
        )
        return

    assert not np.allclose(  # check that distorted structure is not equivalent
        supercell.lattice.parameters, structure.lattice.parameters
    )
    npt.assert_allclose(
        supercell.lattice.parameters, refined_structure.lattice.parameters
    )
    npt.assert_array_almost_equal(
        cluster_subspace.corr_from_structure(structure),
        cluster_subspace.corr_from_structure(refined_structure),
    )


def test_remove_orbits(cluster_subspace, rng):
    subspace = cluster_subspace.copy()  # make copy
    remove_num = rng.integers(2, subspace.num_orbits - 1)
    ids_to_remove = rng.choice(
        range(1, subspace.num_orbits), size=remove_num, replace=False
    )
    subspace.remove_orbits(ids_to_remove)

    assert len(subspace.orbits) == len(cluster_subspace.orbits) - remove_num
    assert subspace.num_orbits == cluster_subspace.num_orbits - remove_num

    for i, orbit in enumerate(cluster_subspace.orbits):
        if i + 1 in ids_to_remove:
            assert orbit not in subspace.orbits
        else:
            assert orbit in subspace.orbits

    corr_inds = [
        i
        for i in range(len(cluster_subspace))
        if cluster_subspace.function_orbit_ids[i] not in ids_to_remove
    ]
    structure = gen_random_structure(subspace.structure, size=2)
    npt.assert_allclose(
        cluster_subspace.corr_from_structure(structure)[corr_inds],
        subspace.corr_from_structure(structure),
    )

    with pytest.raises(ValueError):
        subspace.remove_orbits([-1])
    with pytest.raises(ValueError):
        subspace.remove_orbits([subspace.num_orbits + 1])
    with pytest.raises(ValueError):
        subspace.remove_orbits([0])


def test_remove_corr_functions(cluster_subspace, rng):
    subspace = cluster_subspace.copy()  # make copy
    remove_num = rng.integers(2, len(subspace) - 1)
    ids_to_remove = rng.choice(range(1, len(subspace)), size=remove_num, replace=False)
    subspace.remove_corr_functions(ids_to_remove)

    assert len(subspace) == len(cluster_subspace) - remove_num
    assert (
        subspace.num_corr_functions == cluster_subspace.num_corr_functions - remove_num
    )

    corr_inds = [i for i in range(len(cluster_subspace)) if i not in ids_to_remove]
    structure = gen_random_structure(subspace.structure, size=2)
    npt.assert_allclose(
        cluster_subspace.corr_from_structure(structure)[corr_inds],
        subspace.corr_from_structure(structure),
    )
    with pytest.warns(UserWarning):
        bid = subspace.orbits[-1].bit_id
        ids = list(range(bid, bid + len(subspace.orbits[-1])))
        subspace.remove_corr_functions(ids)


def test_orbit_mappings(cluster_subspace, supercell_matrix):
    # check that all supercell_structure index groups map to the correct
    # primitive cell sites, and check that max distance under supercell
    # structure pbc is less than the max distance without pbc

    # sc_matrix = np.array([[2, 0, 0], [0, 2, 0], [0, 1, 1]])
    supercell_struct = cluster_subspace.structure.copy()
    supercell_struct.make_supercell(supercell_matrix)
    fcoords = np.array(supercell_struct.frac_coords)

    for orb, inds in zip(
        cluster_subspace.orbits,
        cluster_subspace.supercell_orbit_mappings(supercell_matrix),
    ):
        for x in inds:
            pbc_radius = np.max(
                supercell_struct.lattice.get_all_distances(fcoords[x], fcoords[x])
            )
            # primitive cell fractional coordinates
            new_fc = np.dot(fcoords[x], supercell_matrix)
            assert orb.base_cluster.diameter + 1e-7 > pbc_radius
            found = False
            for equiv in orb.clusters:
                if is_coord_subset_pbc(equiv.frac_coords, new_fc, atol=SITE_TOL):
                    found = True
                    break
            assert found

    # check that the matrix was cached
    m_hash = tuple(sorted(tuple(s) for s in supercell_matrix))
    assert cluster_subspace._supercell_orb_inds[
        m_hash
    ] is cluster_subspace.supercell_orbit_mappings(supercell_matrix)


def test_get_aliased_orbits(cluster_subspace, supercell_matrix):
    # Verify that
    # 1) site mappings for aliased orbits are indeed an identical set,
    # 2) each orbit is counted only once,
    aliased_orbs = cluster_subspace.get_aliased_orbits(supercell_matrix)
    sc_orbit_maps = cluster_subspace.supercell_orbit_mappings(supercell_matrix)
    accounted_orbs = []
    for orb_tup in aliased_orbs:
        assert sum(o in accounted_orbs for o in orb_tup) == 0
        accounted_orbs.extend(orb_tup)
        for i, orb_id_i in enumerate(orb_tup[:-1]):
            orbit_map_i = tuple(
                tuple(sorted(inds_i)) for inds_i in sc_orbit_maps[orb_id_i - 1]
            )
            for orb_id_j in orb_tup[i + 1 :]:
                orbit_map_j = tuple(
                    tuple(sorted(inds_j)) for inds_j in sc_orbit_maps[orb_id_j - 1]
                )
                assert set(orbit_map_i) == set(orbit_map_j)


def test_periodicity_and_symmetry(cluster_subspace, supercell_matrix):

    structure = gen_random_structure(cluster_subspace.structure, size=2)
    larger_structure = structure.copy()
    larger_structure.make_supercell(supercell_matrix)

    corr = cluster_subspace.corr_from_structure(structure)

    larger_scmatrix = supercell_matrix @ np.eye(3) * 2
    npt.assert_allclose(
        corr,
        cluster_subspace.corr_from_structure(
            larger_structure, scmatrix=larger_scmatrix
        ),
    )
    # Sometimes when supercell_matrix is very skewed, you should provide it.
    # Structure matcher is not good at finding very off-diagonal sc matrices.

    cm = OrderDisorderElementComparator()
    sm = StructureMatcher(allow_subset=True, comparator=cm, scale=True)

    def apply_operation(op, s):
        """Apply on fractional coordinates only."""
        return Structure(
            s.lattice, [st.species for st in s], op.operate_multi(s.frac_coords)
        )

    for symop in cluster_subspace.symops:
        # Be very careful with symop formats, whether they are fractional space
        # ops or not. By default, SpaceGroupAnalyzer gives fractional format,
        # while PointGroupAnalyzer gives cartesian format!
        prim = cluster_subspace.structure
        prim_op = apply_operation(symop, prim)
        # print("op:", symop)
        # print("prim:\n", prim)
        # print("prim_op:\n", prim_op)
        assert sm.fit(prim, prim_op)

        structure_op = apply_operation(symop, structure)
        npt.assert_allclose(corr, cluster_subspace.corr_from_structure(structure_op))


def test_equality(single_subspace, rng):
    subspace = single_subspace.copy()
    assert subspace == single_subspace
    subspace.change_site_bases("legendre")
    assert subspace == single_subspace
    subspace.remove_orbits(rng.choice(range(1, subspace.num_orbits), 2))
    assert subspace != single_subspace


def test_contains(single_subspace, rng):
    for orbit in single_subspace.orbits:
        assert orbit in single_subspace

    subspace = single_subspace.copy()
    orb_ids = rng.choice(range(1, subspace.num_orbits), 2)
    subspace.remove_orbits(orb_ids)
    for i, orbit in enumerate(single_subspace.orbits):
        if i + 1 in orb_ids:
            assert orbit not in subspace
        else:
            assert orbit in subspace


def test_msonable(cluster_subspace_ewald, rng):
    # force caching some orb indices for a few random structures
    _ = repr(cluster_subspace_ewald)  # can probably do better testing than this...
    _ = str(cluster_subspace_ewald)

    for _ in range(2):
        size = rng.integers(1, 4)
        s = gen_random_structure(cluster_subspace_ewald.structure, size=size)
        _ = cluster_subspace_ewald.corr_from_structure(s)

    assert_msonable(cluster_subspace_ewald)

    subspace = ClusterSubspace.from_dict(cluster_subspace_ewald.as_dict())
    for key in cluster_subspace_ewald._supercell_orb_inds.keys():
        for arr1, arr2 in zip(
            subspace._supercell_orb_inds[key],
            cluster_subspace_ewald._supercell_orb_inds[key],
        ):
            npt.assert_array_equal(arr1, arr2)

    assert len(cluster_subspace_ewald.external_terms) == len(subspace.external_terms)
    npt.assert_allclose(
        subspace.corr_from_structure(s), cluster_subspace_ewald.corr_from_structure(s)
    )


def test_potts_subspace(cluster_subspace, rng):
    potts_subspace = PottsSubspace.from_cutoffs(
        cluster_subspace.structure, cluster_subspace.cutoffs
    )
    assert len(potts_subspace.orbits) == len(cluster_subspace.orbits)

    # check sizes and bits included in each orbit
    for porbit, corbit in zip(potts_subspace.orbits, cluster_subspace.orbits):
        assert len(porbit.site_spaces) == len(corbit.site_spaces)
        assert len(porbit.site_bases) == len(corbit.site_bases)
        assert len(porbit.bit_combos) > len(corbit.bit_combos)

        for i, site_space in enumerate(porbit.site_spaces):
            bits_i = np.concatenate([b[:, i] for b in porbit.bit_combos])
            assert all(j in bits_i for j in site_space.codes)

    # check decorations
    for _ in range(10):
        i = rng.choice(range(1, potts_subspace.num_corr_functions))
        o_id = potts_subspace.function_orbit_ids[i]
        orbit = potts_subspace.orbits[o_id - 1]
        fdeco = potts_subspace.get_function_decoration(i)
        odeco = potts_subspace.get_orbit_decorations(o_id)
        assert fdeco == odeco[i - orbit.bit_id]
        assert all(  # all decorations include valid species
            deco[i] in species
            for deco in fdeco
            for i, species in enumerate(orbit.site_spaces)
        )

    # test removing last bit combo
    potts_subspace1 = PottsSubspace.from_cutoffs(
        cluster_subspace.structure, cluster_subspace.cutoffs, remove_last_cluster=True
    )
    for o1, o2 in zip(potts_subspace.orbits, potts_subspace1.orbits):
        assert len(o1) - 1 == len(o2)

    assert_msonable(potts_subspace)


def test_invert_mapping_table():
    forward = [[], [], [1], [1], [1], [2, 4], [3, 4], [2, 3], [5, 6, 7]]
    backward = [[], [2, 3, 4], [5, 7], [6, 7], [5, 6], [8], [8], [8], []]

    forward_invert = [sorted(sub) for sub in invert_mapping(forward)]
    backward_invert = [sorted(sub) for sub in invert_mapping(backward)]

    assert forward_invert == backward
    assert backward_invert == forward


def test_get_complete_mapping():
    forward = [[], [], [1], [1], [1], [2, 4], [3, 4], [2, 3], [5, 6, 7]]
    backward = [[], [2, 3, 4], [5, 7], [6, 7], [5, 6], [8], [8], [8], []]

    forward_full = [
        [],
        [],
        [1],
        [1],
        [1],
        [1, 2, 4],
        [1, 3, 4],
        [1, 2, 3],
        [1, 2, 3, 4, 5, 6, 7],
    ]
    backward_full = [
        [],
        [2, 3, 4, 5, 6, 7, 8],
        [5, 7, 8],
        [6, 7, 8],
        [5, 6, 8],
        [8],
        [8],
        [8],
        [],
    ]

    forward_comp = [sorted(sub) for sub in get_complete_mapping(forward)]
    backward_comp = [sorted(sub) for sub in get_complete_mapping(backward)]

    assert forward_comp == forward_full
    assert backward_comp == backward_full


# fixed tests for LiCaBr structure
def test_numbers_fixed(single_subspace):
    # Test the total generated orbits, orderings and clusters are
    # as expected.
    assert single_subspace.num_orbits == 27
    assert single_subspace.num_corr_functions == 124
    assert single_subspace.num_clusters == 377


def test_func_orbit_ids_fixed(single_subspace):
    assert len(single_subspace.function_orbit_ids) == 124
    assert len(set(single_subspace.function_orbit_ids)) == 27


def test_function_hierarchy_fixed(single_subspace):
    hierarchy = single_subspace.function_hierarchy()
    assert sorted(hierarchy[0]) == []
    assert sorted(hierarchy[-1]) == [17, 21]
    assert sorted(hierarchy[15]) == []
    assert sorted(hierarchy[35]) == [5, 7, 10]
    assert sorted(hierarchy[55]) == [6, 8, 13]
    assert sorted(hierarchy[75]) == [7, 16, 21]
    assert sorted(hierarchy[95]) == [9, 19]
    assert sorted(hierarchy[115]) == [13, 19, 21]


def test_orbit_hierarchy_fixed(single_subspace):
    hierarchy = single_subspace.orbit_hierarchy()
    assert sorted(hierarchy[0]) == []  # empty
    assert sorted(hierarchy[1]) == []  # point
    assert sorted(hierarchy[3]) == [1, 2]  # distinct site pair
    assert sorted(hierarchy[4]) == [1]  # same site pair
    assert sorted(hierarchy[15]) == [3, 5]  # triplet
    assert sorted(hierarchy[-1]) == [6, 7]


def test_corr_from_structure(single_subspace, rng):
    structure = Structure(
        single_subspace.structure.lattice,
        [
            "Li+",
        ]
        * 2
        + ["Ca+"]
        + ["Br-"],
        single_subspace.structure.frac_coords,
    )
    corr = single_subspace.corr_from_structure(structure)
    assert len(corr) == single_subspace.num_corr_functions + len(
        single_subspace.external_terms
    )
    assert corr[0] == 1

    cs = ClusterSubspace.from_cutoffs(
        single_subspace.structure, {2: 5}, basis="indicator"
    )

    # make an ordered supercell_structure
    s = single_subspace.structure.copy()
    s.make_supercell([2, 1, 1])
    species = ("Li+", "Ca+", "Li+", "Ca+", "Br-", "Br-")
    coords = (
        (0.125, 0.25, 0.25),
        (0.625, 0.25, 0.25),
        (0.375, 0.75, 0.75),
        (0.25, 0.5, 0.5),
        (0, 0, 0),
        (0.5, 0, 0),
    )
    s = Structure(s.lattice, species, coords)
    assert len(cs.corr_from_structure(s)) == 22

    expected = [
        1,
        0.5,
        0.25,
        0,
        0.5,
        0,
        0.375,
        0,
        0.0625,
        0.25,
        0.125,
        0,
        0.25,
        0.125,
        0.125,
        0,
        0,
        0.25,
        0,
        0.125,
        0,
        0.1875,
    ]
    npt.assert_allclose(cs.corr_from_structure(s), expected)

    # Test occu_from_structure
    occu = [
        Vacancy(),
        Species("Li", 1),
        Species("Ca", 1),
        Species("Li", 1),
        Vacancy(),
        Species("Ca", 1),
        Species("Br", -1),
        Species("Br", -1),
    ]
    assert all(s1 == s2 for s1, s2 in zip(occu, cs.occupancy_from_structure(s)))

    # shuffle sites and check correlation still works
    for _ in range(10):
        rng.shuffle(s)
        npt.assert_allclose(cs.corr_from_structure(s), expected)


def test_periodicity(single_subspace):
    # Check to see if a supercell of a smaller structure gives the same corr
    m = np.array([[2, 0, 0], [0, 2, 0], [0, 1, 1]])
    supercell = single_subspace.structure.copy()
    supercell.make_supercell(m)
    s = Structure(
        supercell.lattice,
        ["Ca+", "Li+", "Li+", "Br-", "Br-", "Br-", "Br-"],
        [
            [0.125, 1, 0.25],
            [0.125, 0.5, 0.25],
            [0.375, 0.5, 0.75],
            [0, 0, 0],
            [0, 0.5, 1],
            [0.5, 1, 0],
            [0.5, 0.5, 0],
        ],
    )
    a = single_subspace.corr_from_structure(s)
    s.make_supercell([2, 1, 1])
    b = single_subspace.corr_from_structure(s)
    npt.assert_allclose(a, b)


def _encode_occu(occu, bits):
    return np.array([bit.index(sp) for sp, bit in zip(occu, bits)])


def test_vs_CASM_pairs(single_structure):
    species = [{"Li+": 0.1}] * 3 + ["Br-"]
    coords = ((0.25, 0.25, 0.25), (0.75, 0.75, 0.75), (0.5, 0.5, 0.5), (0, 0, 0))
    structure = Structure(single_structure.lattice, species, coords)
    cs = ClusterSubspace.from_cutoffs(structure, {2: 6}, basis="indicator")

    spaces = get_allowed_species(structure)
    m = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    orbit_list = [
        (
            orbit.bit_id,
            orbit.flat_tensor_indices,
            orbit.flat_correlation_tensors,
            cluster_indices,
        )
        for orbit, cluster_indices in zip(cs.orbits, cs.supercell_orbit_mappings(m))
    ]

    # last two clusters are switched from CASM output (occupancy basis)
    # all_li (ignore casm point term)
    occu = _encode_occu([Species("Li", 1), Species("Li", 1), Species("Li", 1)], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, np.array([1] * 12))

    # all_vacancy
    occu = _encode_occu([Vacancy(), Vacancy(), Vacancy()], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, np.array([1] + [0] * 11))
    # octahedral
    occu = _encode_occu([Vacancy(), Vacancy(), Species("Li", 1)], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, [1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1])
    # tetrahedral
    occu = _encode_occu([Species("Li", 1), Species("Li", 1), Vacancy()], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, [1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0])
    # mixed
    occu = _encode_occu([Species("Li", 1), Vacancy(), Species("Li", 1)], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, [1, 0.5, 1, 0.5, 0, 0.5, 1, 0.5, 0, 0, 0.5, 1])
    # single_tet
    occu = _encode_occu([Species("Li", 1), Vacancy(), Vacancy()], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, [1, 0.5, 0, 0, 0, 0.5, 0, 0, 0, 0, 0.5, 0])


def test_vs_CASM_triplets(single_structure):
    """
    Test vs casm generated correlation with occupancy basis.
    """
    species = [{"Li+": 0.1}] * 3 + ["Br-"]
    coords = ((0.25, 0.25, 0.25), (0.75, 0.75, 0.75), (0.5, 0.5, 0.5), (0, 0, 0))
    structure = Structure(single_structure.lattice, species, coords)
    cs = ClusterSubspace.from_cutoffs(structure, {2: 6, 3: 4.5}, basis="indicator")
    spaces = get_allowed_species(structure)
    m = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

    orbit_list = [
        (
            orbit.bit_id,
            orbit.flat_tensor_indices,
            orbit.flat_correlation_tensors,
            cluster_indices,
        )
        for orbit, cluster_indices in zip(cs.orbits, cs.supercell_orbit_mappings(m))
    ]
    # last two pair terms are switched from CASM output (occupancy basis)
    # all_vacancy (ignore casm point term)
    occu = _encode_occu([Vacancy(), Vacancy(), Vacancy()], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, np.array([1] + [0] * 18))
    # all Li
    occu = _encode_occu([Species("Li", 1), Species("Li", 1), Species("Li", 1)], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, np.array([1] * 19))
    # octahedral
    occu = _encode_occu([Vacancy(), Vacancy(), Species("Li", 1)], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, [1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1])

    # tetrahedral
    occu = _encode_occu([Species("Li", 1), Species("Li", 1), Vacancy()], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(corr, [1, 1, 0, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0])
    # mixed
    occu = _encode_occu([Species("Li", 1), Vacancy(), Species("Li", 1)], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(
        corr,
        [1, 0.5, 1, 0.5, 0, 0.5, 1, 0.5, 0, 0, 0.5, 1, 0, 0, 0.5, 0.5, 0.5, 0.5, 1],
    )
    # single_tet
    occu = _encode_occu([Species("Li", 1), Vacancy(), Vacancy()], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(
        corr, [1, 0.5, 0, 0, 0, 0.5, 0, 0, 0, 0, 0.5, 0, 0, 0, 0, 0, 0.5, 0.5, 0]
    )


def test_vs_CASM_multicomp(single_structure):
    cs = ClusterSubspace.from_cutoffs(single_structure, {2: 5}, basis="indicator")
    m = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    spaces = get_allowed_species(single_structure)
    orbit_list = [
        (
            orbit.bit_id,
            orbit.flat_tensor_indices,
            orbit.flat_correlation_tensors,
            cluster_indices,
        )
        for orbit, cluster_indices in zip(cs.orbits, cs.supercell_orbit_mappings(m))
    ]
    # mixed
    occu = _encode_occu([Vacancy(), Species("Li", 1), Species("Li", 1)], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(
        corr, [1, 0.5, 0, 1, 0, 0.5, 0, 0, 0, 0, 0, 0, 0.5, 0, 0, 1, 0, 0, 0.5, 0, 0, 0]
    )

    # Li-tet Ca-oct
    occu = _encode_occu([Vacancy(), Species("Li", 1), Species("Ca", 1)], spaces)
    corr = corr_from_occupancy(occu, cs.num_corr_functions, orbit_list)
    npt.assert_allclose(
        corr, [1, 0.5, 0, 0, 1, 0, 0.5, 0, 0, 0, 0, 0, 0.5, 0, 0, 0, 0, 1, 0, 0.5, 0, 0]
    )



"""
Implementation of a Canonical Ensemble class.

These are used when running Monte Carlo simulations for systems
with a fixed number of sites and fixed concentration of species.
"""

__author__ = "Luis Barroso-Luque"


from monty.json import MSONable

from smol.moca.ensemble.base import Ensemble
from smol.moca.processor.base import Processor
from smol.moca.sublattice import Sublattice


class CanonicalEnsemble(Ensemble, MSONable):
    """Canonical Ensemble class to run Monte Carlo Simulations."""

    valid_mcmc_steps = ("swap",)

    @property
    def natural_parameters(self):
        """Get the vector of exponential parameters."""
        return self.processor.coefs

    def compute_feature_vector(self, occupancy):
        """Compute the feature vector for a given occupancy.

        In the canonical case it is just the feature vector from the underlying
        processor.

        Args:
            occupancy (ndarray):
                encoded occupancy string

        Returns:
            ndarray: feature vector
        """
        return self.processor.compute_feature_vector(occupancy)

    def compute_feature_vector_change(self, occupancy, step):
        """Compute the change in the feature vector from a given step.

        Args:
            occupancy (ndarray):
                encoded occupancy string.
            step (list of tuple):
                A sequence of flips as given by MCUsher.propose_step

        Returns:
            ndarray: difference in feature vector
        """
        return self.processor.compute_feature_vector_change(occupancy, step)

    @classmethod
    def from_dict(cls, d):
        """Instantiate a CanonicalEnsemble from dict representation.

        Args:
            d (dict):
                dictionary representation.
        Returns:
            CanonicalEnsemble
        """
        return cls(
            Processor.from_dict(d["processor"]),
            [Sublattice.from_dict(s) for s in d["sublattices"]],
        )



import numpy as np
import numpy.testing as npt
import pytest

from smol.moca import Sampler
from smol.moca.sampler.kernel import Metropolis
from smol.moca.sampler.mcusher import Flip, Swap
from tests.utils import gen_random_occupancy

TEMPERATURE = 5000


@pytest.fixture(params=[1, 5])
def sampler(ensemble, rng, request):
    sampler = Sampler.from_ensemble(
        ensemble, temperature=TEMPERATURE, seed=rng, nwalkers=request.param
    )
    # fix this additional attribute to sampler to access in gen occus for tests
    sampler.num_sites = ensemble.num_sites
    return sampler


def test_from_ensemble(sampler):
    if "Canonical" in sampler.samples.metadata["name"]:
        assert isinstance(sampler.mckernel._usher, Swap)
    else:
        assert isinstance(sampler.mckernel._usher, Flip)
    assert isinstance(sampler.mckernel, Metropolis)


@pytest.mark.parametrize("thin", (1, 10))
def test_sample(sampler, thin):
    occu = np.vstack(
        [
            gen_random_occupancy(sampler.mckernel._usher.sublattices)
            for _ in range(sampler.samples.shape[0])
        ]
    )
    steps = 1000
    samples = [state for state in sampler.sample(1000, occu, thin_by=thin)]
    assert len(samples) == steps // thin

    it = sampler.sample(43, occu, thin_by=7)
    with pytest.warns(RuntimeWarning):
        next(it)


# TODO efficiency is sometimes =0 and so fails
@pytest.mark.parametrize("thin", (1, 10))
def test_run(sampler, thin):
    occu = np.vstack(
        [
            gen_random_occupancy(sampler.mckernel._usher.sublattices)
            for _ in range(sampler.samples.shape[0])
        ]
    )
    steps = 1000
    sampler.run(steps, occu, thin_by=thin)
    assert len(sampler.samples) == steps // thin
    assert 0 <= sampler.efficiency() <= 1
    sampler.clear_samples()


def test_anneal(sampler):
    temperatures = np.linspace(2000, 500, 5)
    occu = np.vstack(
        [
            gen_random_occupancy(sampler.mckernel._usher.sublattices)
            for _ in range(sampler.samples.shape[0])
        ]
    )
    steps = 100
    sampler.anneal(temperatures, steps, occu)
    expected = []
    for T in temperatures:
        expected += (
            steps
            * sampler.samples.shape[0]
            * [
                T,
            ]
        )
    npt.assert_array_equal(sampler.samples.get_trace_value("temperature"), expected)
    # test temp error
    with pytest.raises(ValueError):
        sampler.anneal([100, 200], steps)


# TODO test run sgensembles at high temp
"""
# test some high temp high potential values
steps = 10000
chem_pots = {'Na+': 100.0, 'Cl-': 0.0}
self.msgensemble.chemical_potentials = chem_pots
expected = {'Na+': 1.0, 'Cl-': 0.0}
sampler_m.run(steps, self.occu)
comp = sampler_m.samples.mean_composition()
for sp in expected.keys():
    self.assertAlmostEqual(expected[sp], comp[sp], places=2)
sampler_m.clear_samples()

chem_pots = {'Na+': -100.0, 'Cl-': 0.0}
self.msgensemble.chemical_potentials = chem_pots
expected = {'Na+': 0.0, 'Cl-': 1.0}
sampler_m.run(steps, self.occu)
comp = sampler_m.samples.mean_composition()
for sp in expected.keys():
    self.assertAlmostEqual(expected[sp], comp[sp], places=2)
sampler_m.clear_samples()

self.fsgensemble.temperature = 1E9  # go real high to be uniform
sampler_f.run(steps, self.occu)
expected = {'Na+': 0.5, 'Cl-': 0.5}
comp = sampler_f.samples.mean_composition()
for sp in expected.keys():
    self.assertAlmostEqual(expected[sp], comp[sp], places=1)
"""


def test_reshape_occu(ensemble):
    sampler = Sampler.from_ensemble(ensemble, temperature=TEMPERATURE)
    occu = gen_random_occupancy(ensemble.sublattices)
    assert sampler._reshape_occu(occu).shape == (1, len(occu))



import numpy as np
import numpy.testing as npt
import pytest
from sklearn.linear_model import LinearRegression, Ridge

from smol.cofe import ClusterExpansion, RegressionData
from tests.utils import assert_msonable, gen_random_structure


@pytest.fixture(scope="module")
def cluster_expansion(cluster_subspace, rng):
    reg = Ridge(alpha=1e-8, fit_intercept=False)
    n = rng.integers(50, 100)
    feat_matrix = np.empty((n, len(cluster_subspace)))
    structures = []
    for i in range(n):
        structure = gen_random_structure(
            cluster_subspace.structure, size=rng.integers(2, 5)
        )
        structures.append(structure)
        feat_matrix[i] = cluster_subspace.corr_from_structure(structure)

    prop_vec = -5 * rng.random(n)
    reg.fit(feat_matrix, prop_vec)
    reg_data = RegressionData.from_sklearn(
        reg, feature_matrix=feat_matrix, property_vector=prop_vec
    )
    expansion = ClusterExpansion(cluster_subspace, reg.coef_, reg_data)
    # bind the structures to the expansion willy nilly
    expansion.structures = structures
    return expansion


def test_regression_data(cluster_subspace, rng):
    reg = LinearRegression(fit_intercept=False)
    n = rng.integers(10, 100)
    feat_matrix = rng.random((n, len(cluster_subspace)))
    prop_vec = rng.random(n)
    reg_data = RegressionData.from_sklearn(
        reg, feature_matrix=feat_matrix, property_vector=prop_vec
    )
    coeffs = rng.random(len(cluster_subspace))
    expansion = ClusterExpansion(cluster_subspace, coeffs, reg_data)
    assert reg_data.estimator_name == reg.__class__.__name__
    assert reg_data.module == reg.__module__
    assert reg_data.parameters == reg.get_params()
    assert_msonable(expansion)


def test_predict(cluster_expansion, rng):
    subspace = cluster_expansion.cluster_subspace

    prim = cluster_expansion.structure
    scmatrix = np.eye(3, dtype=int) * 3
    scmatrix[0, 1] = 2  # Intentionally made less symmetric
    scmatrix[1, 2] = 1
    N = np.abs(np.linalg.det(scmatrix))
    pool = [gen_random_structure(prim, scmatrix) for _ in range(100)]
    feature_matrix = np.array(
        [
            subspace.corr_from_structure(s, scmatrix=scmatrix, normalized=True)
            for s in pool
        ]
    )

    comps = [s.composition for s in pool]
    all_species = list({b for c in comps for b in c.keys()})
    mus = rng.random(len(all_species))

    def get_energy(structure, species, chempots):
        return np.dot(chempots, [structure.composition[sp] for sp in species])

    energies = np.array([get_energy(s, all_species, mus) for s in pool]) / N
    reg = LinearRegression(fit_intercept=False)
    reg.fit(feature_matrix, energies)
    coefs = reg.coef_
    expansion_new = ClusterExpansion(subspace, coefs)
    # Why don't we add a "scmatrix" option into ClusterExpansion.predict?
    # This will make it safer to structure skew, because pymatgen can't seem
    # to figure out highly skewed supercell matrix correctly.
    energies_pred = np.array(
        [expansion_new.predict(s, scmatrix=scmatrix, normalize=True) for s in pool]
    )
    np.testing.assert_almost_equal(energies, energies_pred, decimal=6)


def test_prune(cluster_expansion):
    expansion = cluster_expansion.copy()
    thresh = 1e-2
    expansion.prune(threshold=thresh)
    ids = [i for i, coef in enumerate(cluster_expansion.coefs) if abs(coef) >= thresh]
    new_coefs = cluster_expansion.coefs[ids]
    new_eci = cluster_expansion.eci[ids]
    assert len(expansion.coefs) == len(new_coefs)
    npt.assert_array_equal(new_eci, expansion.eci)
    npt.assert_array_equal(new_coefs, expansion.coefs)
    assert len(expansion.cluster_subspace) == len(new_coefs)
    assert len(expansion.eci_orbit_ids) == len(new_coefs)
    pruned_feat_matrix = cluster_expansion._feat_matrix[:, ids]
    npt.assert_array_equal(expansion._feat_matrix, pruned_feat_matrix)
    # check that recomputing features produces whats expected
    new_feature_matrix = np.array(
        [
            expansion.cluster_subspace.corr_from_structure(s)
            for s in cluster_expansion.structures
        ]
    )
    npt.assert_array_equal(new_feature_matrix, pruned_feat_matrix)
    # check new predictions
    preds = [expansion.predict(s, normalize=True) for s in cluster_expansion.structures]
    npt.assert_allclose(preds, np.dot(pruned_feat_matrix, new_coefs))


def test_msonable(cluster_expansion):
    _ = str(cluster_expansion)
    d = cluster_expansion.as_dict()
    ce1 = ClusterExpansion.from_dict(d)
    npt.assert_array_equal(cluster_expansion.coefs, ce1.coefs)
    # change this to just use assert_msonable
    assert_msonable(cluster_expansion)



"""Implementation of StructureWrangler.

And additional wrangling functions to preprocess and filter fitting data.
"""

from .select import composition_select, full_row_rank_select, gaussian_select
from .tools import (
    max_ewald_energy_indices,
    unique_corr_vector_indices,
    weights_energy_above_composition,
    weights_energy_above_hull,
)

__all__ = [
    "unique_corr_vector_indices",
    "max_ewald_energy_indices",
    "weights_energy_above_composition",
    "weights_energy_above_hull",
    "full_row_rank_select",
    "gaussian_select",
    "composition_select",
]



"""Abstract base class for Monte Carlo Ensembles."""

__author__ = "Luis Barroso-Luque"

from abc import ABC, abstractmethod

from smol.moca.processor import (
    ClusterExpansionProcessor,
    CompositeProcessor,
    EwaldProcessor,
)


class Ensemble(ABC):
    """Abstract base class for Monte Carlo Ensembles.

    Attributes:
        num_energy_coefs (int):
            Number of coefficients in the natural parameters array that
            correspond to energy only.
        thermo_boundaries (dict):
            Dictionary with corresponding thermodynamic boundaries, i.e.
            chemical potentials or fugacity fractions. This is kept only for
            descriptive purposes.
        valid_mcmc_steps (list of str):
            List of the valid MCMC steps that can be used to sample the
            ensemble in MCMC.
    """

    valid_mcmc_steps = None  # add this in derived classes

    def __init__(self, processor, sublattices=None):
        """Initialize class instance.

        Args:
            processor (Processor):
                a processor that can compute the change in property given
                a set of flips.
            sublattices (list of Sublattice): optional
                list of Sublattice objects representing sites in the processor
                supercell with same site spaces.
        """
        if sublattices is None:
            sublattices = processor.get_sublattices()
        self.num_energy_coefs = len(processor.coefs)
        self.thermo_boundaries = {}  # not pretty way to save general info
        self._processor = processor
        self._sublattices = sublattices

    @classmethod
    def from_cluster_expansion(cls, cluster_expansion, supercell_matrix, **kwargs):
        """Initialize an ensemble from a cluster expansion.

        Convenience constructor to instantiate an ensemble. This will take
        care of initializing the correct processor based on the
        ClusterExpansion.

        Args:
            cluster_expansion (ClusterExpansion):
                A cluster expansion object.
            supercell_matrix (ndarray):
                Supercell matrix defining the system size.
            **kwargs:
                Keyword arguments to pass to ensemble constructor. Such as
                sublattices, sublattice_probabilities, chemical_potentials,
                fugacity_fractions.

        Returns:
            Ensemble
        """
        if len(cluster_expansion.cluster_subspace.external_terms) > 0:
            processor = CompositeProcessor(
                cluster_expansion.cluster_subspace, supercell_matrix
            )
            ceprocessor = ClusterExpansionProcessor(
                cluster_expansion.cluster_subspace,
                supercell_matrix,
                cluster_expansion.coefs[:-1],
            )
            processor.add_processor(ceprocessor)
            # at some point determine term and spinup processor maybe with a
            # factory, if we ever implement more external terms.
            ewald_term = cluster_expansion.cluster_subspace.external_terms[0]
            ewprocessor = EwaldProcessor(
                cluster_expansion.cluster_subspace,
                supercell_matrix,
                ewald_term=ewald_term,
                coefficient=cluster_expansion.coefs[-1],
            )
            processor.add_processor(ewprocessor)
        else:
            processor = ClusterExpansionProcessor(
                cluster_expansion.cluster_subspace,
                supercell_matrix,
                cluster_expansion.coefs,
            )
        return cls(processor, **kwargs)

    @property
    def num_sites(self):
        """Get the total number of sites in the supercell."""
        return self.processor.num_sites

    @property
    def system_size(self):
        """Get size of supercell in number of prims."""
        return self.processor.size

    @property
    def processor(self):
        """Get the ensemble processor."""
        return self._processor

    # TODO make a setter for these that checks sublattices are correct and
    #  all sites are included.
    @property
    def sublattices(self):
        """Get list of Sublattices included in ensemble."""
        return self._sublattices

    @property
    def active_sublattices(self):
        """Get list of active sub-lattices."""
        return [s for s in self.sublattices if s.is_active]

    @property
    def restricted_sites(self):
        """Get indices of all restricted sites."""
        sites = []
        for sublattice in self.sublattices:
            sites += sublattice.restricted_sites
        return sites

    @property
    def species(self):
        """Species on active sublattices.

        These are minimal species required in setting chemical potentials.
        """
        return list(
            {sp for sublatt in self.active_sublattices for sp in sublatt.site_space}
        )

    def split_sublattice_by_species(self, sublattice_id, occu, species_in_partitions):
        """Split a sub-lattice in system by its occupied species.

        An example use case might be simulating topotactic Li extraction
        and insertion, where we want to consider Li/Vac, TM and O as
        different sub-lattices that can not be mixed by swapping.

        Args:
            sublattice_id (int):
                The index of sub-lattice to split in self.sublattices.
            occu (np.ndarray[int]):
                An occupancy array to reference with.
            species_in_partitions (List[List[int|Species|Vacancy|Element|str]]):
                Each sub-list contains a few species or encodings of species in
                the site space to be grouped as a new sub-lattice, namely,
                sites with occu[sites] == specie in the sub-list, will be
                used to initialize a new sub-lattice.
                Sub-lists will be pre-sorted to ascending order.
        """
        splits = self.sublattices[sublattice_id].split_by_species(
            occu, species_in_partitions
        )
        self._sublattices = (
            self._sublattices[:sublattice_id]
            + splits
            + self._sublattices[sublattice_id + 1 :]
        )

    @property
    @abstractmethod
    def natural_parameters(self):
        """Get the vector of natural parameters.

        The natural parameters correspond to the fit coefficients of the
        underlying processor plus any additional terms involved in the Legendre
        transformation corresponding to the ensemble.
        """
        return

    @abstractmethod
    def compute_feature_vector(self, occupancy):
        """Compute the feature vector for a given occupancy.

        The feature vector includes the necessary features required to compute
        the exponent determining the relative probability for the given
        occupancy (i.e. a generalized enthalpy). The feature vector for
        ensembles represents the sufficient statistics.

        For a cluster expansion, the feature vector is the
        correlation vector x system size

        Args:
            occupancy (ndarray):
                encoded occupancy string

        Returns:
            ndarray: feature vector
        """
        return

    @abstractmethod
    def compute_feature_vector_change(self, occupancy, step):
        """Compute the change in the feature vector from a given step.

        Args:
            occupancy (ndarray):
                encoded occupancy string.
            step (list of tuple):
                a sequence of flips given by MCUsher.propose_step

        Returns:
            ndarray: difference in feature vector
        """
        return

    def restrict_sites(self, sites):
        """Restricts (freezes) the given sites.

        This will exclude those sites from being flipped during a Monte Carlo
        run. If some of the given indices refer to inactive sites, there will
        be no effect.

        Args:
            sites (Sequence):
                indices of sites in the occupancy string to restrict.
        """
        for sublattice in self.sublattices:
            sublattice.restrict_sites(sites)

    def reset_restricted_sites(self):
        """Unfreeze all previously restricted sites."""
        for sublattice in self.sublattices:
            sublattice.reset_restricted_sites()

    def as_dict(self):
        """Get Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        ensemble_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "thermo_boundaries": self.thermo_boundaries,
            "processor": self._processor.as_dict(),
            "sublattices": [s.as_dict() for s in self._sublattices],
        }
        return ensemble_d



from itertools import product

import numpy as np
import numpy.testing as npt
import pytest

from smol.constants import kB
from smol.moca.sampler.bias import FugacityBias
from smol.moca.sampler.kernel import (
    ALL_MCUSHERS,
    Metropolis,
    StepTrace,
    ThermalKernel,
    Trace,
    UniformlyRandom,
)
from smol.moca.sampler.mcusher import Flip, Swap
from tests.utils import gen_random_occupancy

kernels = [UniformlyRandom, Metropolis]
ushers = ALL_MCUSHERS


@pytest.fixture(params=product(kernels, ushers), scope="module")
def mckernel(ensemble, request):
    kwargs = {}
    kernel_class, step_type = request.param
    if issubclass(kernel_class, ThermalKernel):
        kwargs["temperature"] = 5000
    kernel = kernel_class(ensemble, step_type=step_type, **kwargs)
    return kernel


@pytest.fixture(params=product(kernels, ushers), scope="module")
def mckernel_bias(ensemble, request):
    kwargs = {}
    kernel_class, step_type = request.param
    if issubclass(kernel_class, ThermalKernel):
        kwargs["temperature"] = 5000
    kernel = kernel_class(ensemble, step_type=step_type, **kwargs)
    kernel.bias = FugacityBias(kernel.mcusher.sublattices)
    return kernel


@pytest.mark.parametrize("step_type, mcusher", [("swap", Swap), ("flip", Flip)])
def test_constructor(ensemble, step_type, mcusher):
    kernel = Metropolis(ensemble, step_type=step_type, temperature=500)
    assert isinstance(kernel._usher, mcusher)
    assert isinstance(kernel.trace, StepTrace)
    assert "temperature" in kernel.trace.names
    kernel.bias = FugacityBias(kernel.mcusher.sublattices)
    assert "bias" in kernel.trace.delta_trace.names


def test_trace(rng):
    trace = Trace(first=np.ones(10), second=np.zeros(10))
    assert all(isinstance(val, np.ndarray) for _, val in trace.items())

    trace.third = rng.random(10)
    assert all(isinstance(val, np.ndarray) for _, val in trace.items())
    names = ["first", "second", "third"]
    assert all(name in names for name in trace.names)

    with pytest.raises(TypeError):
        trace.fourth = "blabla"
        Trace(one=np.zeros(40), two=66)

    steptrace = StepTrace(one=np.zeros(10))
    assert isinstance(steptrace.delta_trace, Trace)
    with pytest.raises(ValueError):
        steptrace.delta_trace = np.ones(10)
        StepTrace(delta_trace=np.ones(10))

    # check saving
    assert trace.as_dict() == trace.__dict__
    steptrace_d = steptrace.__dict__.copy()
    steptrace_d["delta_trace"] = steptrace_d["delta_trace"].__dict__.copy()
    assert steptrace.as_dict() == steptrace_d


def test_single_step(mckernel):
    occu_ = gen_random_occupancy(mckernel._usher.sublattices)
    for _ in range(20):
        trace = mckernel.single_step(occu_.copy())
        if trace.accepted:
            assert not np.array_equal(trace.occupancy, occu_)
        else:
            npt.assert_array_equal(trace.occupancy, occu_)


def test_single_step_bias(mckernel_bias):
    occu = gen_random_occupancy(mckernel_bias._usher.sublattices)
    for _ in range(20):
        trace = mckernel_bias.single_step(occu.copy())
        # assert delta bias is there and recorded
        assert isinstance(trace.delta_trace.bias, np.ndarray)
        print(trace.delta_trace.bias)
        assert len(trace.delta_trace.bias.shape) == 0  # 0 dimensional
        if trace.accepted:
            assert not np.array_equal(trace.occupancy, occu)
        else:
            npt.assert_array_equal(trace.occupancy, occu)


def test_temperature_setter(single_canonical_ensemble):
    metropolis_kernel = Metropolis(
        single_canonical_ensemble, step_type="flip", temperature=500
    )
    assert metropolis_kernel.beta == 1 / (kB * metropolis_kernel.temperature)
    metropolis_kernel.temperature = 500
    assert metropolis_kernel.beta == 1 / (kB * 500)



from copy import deepcopy

import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.entries.computed_entries import ComputedStructureEntry

from smol.cofe import StructureWrangler
from smol.cofe.extern import EwaldTerm
from tests.utils import assert_msonable, gen_fake_training_data, gen_random_structure


def test_add_data(structure_wrangler, rng):
    for entry in gen_fake_training_data(
        structure_wrangler.cluster_subspace.structure, rng=rng
    ):
        structure_wrangler.add_entry(entry, weights={"random": 2.0})
    struct = gen_random_structure(structure_wrangler.cluster_subspace.structure)
    energy = -len(struct) * rng.random()
    structure_wrangler.add_entry(
        ComputedStructureEntry(
            struct,
            energy,
            data={"weights": {"random": 3.0}, "properties": {"random": rng.random()}},
        )
    )

    assert all(w == 2.0 for w in structure_wrangler.get_weights("random")[:-1])
    assert (
        len(structure_wrangler.get_weights("random"))
        == structure_wrangler.num_structures
    )
    assert structure_wrangler.get_weights("random")[-1] == 3.0
    assert structure_wrangler.available_properties == ["random"]
    assert structure_wrangler.available_weights == ["random"]

    entries = structure_wrangler._entries
    structure_wrangler.remove_all_data()
    assert len(structure_wrangler.structures) == 0

    # Test passing supercell matrices
    for entry in entries:
        structure_wrangler.add_entry(
            entry,
            entry.data["properties"],
            weights=entry.data["weights"],
            supercell_matrix=entry.data["supercell_matrix"],
        )
    assert len(structure_wrangler.structures) == len(entries)
    structure_wrangler.remove_all_data()
    assert len(structure_wrangler.structures) == 0

    # Test passing site mappings
    for entry in entries:
        structure_wrangler.add_entry(
            entry,
            entry.data["properties"],
            weights=entry.data["weights"],
            site_mapping=entry.data["site_mapping"],
        )

    assert len(structure_wrangler._entries) == len(entries)
    structure_wrangler.remove_all_data()
    # test passing both
    for entry in entries:
        structure_wrangler.add_entry(
            entry,
            entry.data["properties"],
            weights=entry.data["weights"],
            supercell_matrix=entry.data["supercell_matrix"],
            site_mapping=entry.data["site_mapping"],
        )
    assert len(structure_wrangler.structures) == len(entries)

    # Add more properties to test removal
    structure_wrangler.add_properties(
        "normalized", structure_wrangler.get_property_vector("energy", normalize=True)
    )
    structure_wrangler.add_properties(
        "normalized1", structure_wrangler.get_property_vector("energy", normalize=True)
    )
    assert all(
        prop in ["normalized_energy", "normalized", "normalized1", "random"]
        for prop in structure_wrangler.available_properties
    )
    structure_wrangler.remove_properties(
        "normalized_energy", "normalized", "normalized1"
    )
    assert structure_wrangler.available_properties == ["random"]

    # heavily distorted structure
    struct = struct.copy()
    struct.apply_strain(0.2)
    entry = ComputedStructureEntry(struct, energy)
    with pytest.raises(Exception):
        structure_wrangler.add_entry(
            entry,
            properties={"random": rng.random()},
            weights={"random": 3.0},
            raise_failed=True,
        )

    with pytest.warns(UserWarning):
        structure_wrangler.add_entry(
            entry,
            properties={"random": rng.random()},
            weights={"random": 3.0},
            raise_failed=False,
        )

    with pytest.raises(AttributeError):
        structure_wrangler.add_properties("test", structure_wrangler.sizes[:-2])

    structure_wrangler.add_properties(
        "normalized_energy",
        structure_wrangler.get_property_vector("energy", normalize=True),
    )

    entry = deepcopy(
        structure_wrangler.entries[rng.choice(range(structure_wrangler.num_structures))]
    )

    with pytest.raises(ValueError):
        entry.data["properties"].update({"foo": 10})
        structure_wrangler.add_entry(
            ComputedStructureEntry(entry.structure, entry.energy),
            entry.data["properties"],
            weights=entry.data["weights"],
            supercell_matrix=entry.data["supercell_matrix"],
            site_mapping=entry.data["site_mapping"],
        )
    with pytest.raises(ValueError):
        entry.data["properties"] = {}
        structure_wrangler.add_entry(
            ComputedStructureEntry(entry.structure, entry.energy),
            entry.data["properties"],
            weights=entry.data["weights"],
            supercell_matrix=entry.data["supercell_matrix"],
            site_mapping=entry.data["site_mapping"],
        )
    with pytest.raises(ValueError):
        entry.data["weights"].update({"foo": 10})
        structure_wrangler.add_entry(
            ComputedStructureEntry(entry.structure, entry.energy),
            entry.data["properties"],
            weights=entry.data["weights"],
            supercell_matrix=entry.data["supercell_matrix"],
            site_mapping=entry.data["site_mapping"],
        )
    with pytest.raises(ValueError):
        _ = entry.data["weights"].pop("random")
        structure_wrangler.add_entry(
            ComputedStructureEntry(entry.structure, entry.energy),
            entry.data["properties"],
            weights=entry.data["weights"],
            supercell_matrix=entry.data["supercell_matrix"],
            site_mapping=entry.data["site_mapping"],
        )

    with pytest.warns(RuntimeWarning):
        structure_wrangler.remove_properties("blab")


def test_add_weights(structure_wrangler, rng):
    sc_matrices = structure_wrangler.supercell_matrices
    num_structs = structure_wrangler.num_structures
    structures = structure_wrangler.structures
    energies = structure_wrangler.get_property_vector("energy")
    weights = rng.random(num_structs)
    structure_wrangler.add_weights("comp", weights)
    structure_wrangler.remove_all_data()
    assert structure_wrangler.num_structures == 0
    for struct, energy, weight, matrix in zip(
        structures, energies, weights, sc_matrices
    ):
        structure_wrangler.add_entry(
            ComputedStructureEntry(struct, energy),
            weights={"comp": weight},
            supercell_matrix=matrix,
        )
    assert num_structs == structure_wrangler.num_structures
    npt.assert_array_almost_equal(
        weights, structure_wrangler.get_weights("comp"), decimal=9
    )
    with pytest.raises(AttributeError):
        structure_wrangler.add_weights("test", weights[:-2])


def test_append_entries(structure_wrangler, rng):
    entries = structure_wrangler._entries
    structure_wrangler.remove_all_data()
    assert len(structure_wrangler.structures) == 0

    processed_entries = []
    for entry in entries:
        processed_entries.append(
            structure_wrangler.process_entry(
                ComputedStructureEntry(entry.structure, entry.energy),
                properties={"random": 1.0},
                weights=entry.data["weights"],
                supercell_matrix=entry.data["supercell_matrix"],
                site_mapping=entry.data["site_mapping"],
            )
        )

    structure_wrangler.append_entries(processed_entries[:-2])
    structure_wrangler.append_entries(processed_entries[-2:])
    assert len(structure_wrangler.entries) == len(entries)
    entry = deepcopy(
        structure_wrangler.entries[rng.choice(range(structure_wrangler.num_structures))]
    )

    with pytest.raises(ValueError):
        entry.data["properties"].update({"foo": 10})
        structure_wrangler.append_entries([entry])
    with pytest.raises(ValueError):
        _ = entry.data["properties"].pop("random")
        structure_wrangler.append_entries([entry])
    with pytest.raises(ValueError):
        entry.data["weights"].update({"foo": 10})
        structure_wrangler.append_entries([entry])
    with pytest.raises(ValueError):
        _ = entry.data["weights"].pop("random")
        structure_wrangler.append_entries([entry])


def test_data_indices(structure_wrangler, rng):
    test = rng.choice(range(structure_wrangler.num_structures), 5)
    train = np.setdiff1d(range(structure_wrangler.num_structures), test)
    structure_wrangler.add_data_indices("test", test)
    structure_wrangler.add_data_indices("train", train)
    assert all(key in structure_wrangler.available_indices for key in ["test", "train"])
    with pytest.raises(ValueError):
        structure_wrangler.add_data_indices(
            "bla",
            [
                structure_wrangler.num_structures,
            ],
        )

    with pytest.raises(TypeError):
        structure_wrangler.add_data_indices("foo", 77)


def test_properties(structure_wrangler):
    assert structure_wrangler.feature_matrix.shape == (
        structure_wrangler.num_structures,
        structure_wrangler.cluster_subspace.num_corr_functions,
    )
    assert (
        len(structure_wrangler.occupancy_strings) == structure_wrangler.num_structures
    )
    num_prim_sites = len(structure_wrangler.cluster_subspace.structure)
    for struct, occu, size in zip(
        structure_wrangler.structures,
        structure_wrangler.occupancy_strings,
        structure_wrangler.sizes,
    ):
        assert len(struct) <= len(occu)  # < with vacancies
        assert size * num_prim_sites == len(occu)


def test_remove_entry(structure_wrangler, rng):
    total = len(structure_wrangler.structures)
    entry = structure_wrangler.entries[rng.integers(0, total)]
    structure_wrangler.remove_entry(entry)
    assert len(structure_wrangler.structures) == total - 1
    with pytest.warns(RuntimeWarning):
        structure_wrangler.remove_entry(entry)


def test_update_features(structure_wrangler):
    shape = structure_wrangler.feature_matrix.shape
    structure_wrangler.cluster_subspace.add_external_term(EwaldTerm())
    structure_wrangler.update_features()
    assert shape[1] + 1 == structure_wrangler.feature_matrix.shape[1]
    structure_wrangler.update_features()


def test_get_gram_matrix(structure_wrangler, rng):
    G = structure_wrangler.get_gram_matrix()
    assert G.shape == 2 * (structure_wrangler.num_features,)
    npt.assert_array_equal(G, G.T)
    npt.assert_array_almost_equal(np.ones(G.shape[0]), G.diagonal())

    rows = rng.choice(
        range(structure_wrangler.num_structures), structure_wrangler.num_structures - 2
    )
    cols = rng.choice(
        range(structure_wrangler.num_features), structure_wrangler.num_features - 4
    )
    G = structure_wrangler.get_gram_matrix(rows=rows, cols=cols, normalize=False)
    assert G.shape == 2 * (structure_wrangler.num_features - 4,)
    npt.assert_array_equal(G, G.T)
    assert not np.allclose(np.ones(G.shape[0]), G.diagonal())


def test_get_similarity_matrix(structure_wrangler, rng):
    S = structure_wrangler.get_similarity_matrix()
    assert S.shape == 2 * (structure_wrangler.num_features,)
    npt.assert_array_equal(S, S.T)
    npt.assert_array_equal(S.diagonal(), np.ones(S.shape[0]))
    rows = rng.choice(
        range(structure_wrangler.num_structures), structure_wrangler.num_structures - 2
    )
    cols = rng.choice(
        range(structure_wrangler.num_features), structure_wrangler.num_features - 4
    )

    S = structure_wrangler.get_similarity_matrix(rows=rows, cols=cols)
    assert S.shape == 2 * (structure_wrangler.num_features - 4,)
    npt.assert_array_equal(S, S.T)
    npt.assert_array_equal(np.ones(S.shape[0]), S.diagonal())


def test_matrix_properties(structure_wrangler, rng):
    assert structure_wrangler.get_condition_number() >= 1
    rows = rng.choice(
        range(structure_wrangler.num_structures), structure_wrangler.num_structures // 3
    )
    cols = rng.choice(
        range(structure_wrangler.num_features), structure_wrangler.num_features // 3
    )
    assert structure_wrangler.get_condition_number() >= 1
    assert structure_wrangler.get_condition_number(rows, cols) >= 1
    assert structure_wrangler.get_feature_matrix_rank(
        rows, cols
    ) >= structure_wrangler.get_feature_matrix_rank(rows=rows, cols=cols[:-3])


def test_get_orbit_rank(structure_wrangler, rng):
    for _ in range(10):
        oid = rng.choice(range(1, len(structure_wrangler.cluster_subspace.orbits) + 1))
        orb_size = structure_wrangler.cluster_subspace.orbits[oid - 1]
        assert structure_wrangler.get_feature_matrix_orbit_rank(oid) <= len(orb_size)


def test_get_duplicate_corr_inds(structure_wrangler, rng):
    ind = rng.integers(structure_wrangler.num_structures)
    dup_item = deepcopy(structure_wrangler.entries[ind])
    with pytest.warns(UserWarning):
        structure_wrangler.add_entry(dup_item, dup_item.data["properties"])

    assert [
        ind,
        structure_wrangler.num_structures - 1,
    ] in structure_wrangler.get_duplicate_corr_indices()


def test_get_matching_corr_duplicate_inds(structure_wrangler, rng):
    ind = rng.integers(structure_wrangler.num_structures)
    dup_entry = deepcopy(structure_wrangler.entries[ind])
    ind2 = rng.integers(structure_wrangler.num_structures)
    # change the structure for this one:
    struct = rng.choice(
        [s for s in structure_wrangler.structures if s != dup_entry.structure]
    )
    dup_entry2 = structure_wrangler.process_entry(
        ComputedStructureEntry(struct, structure_wrangler.entries[ind2].energy),
        properties=structure_wrangler.entries[ind2].data["properties"],
        weights=structure_wrangler.entries[ind2].data["weights"],
    )
    structure_wrangler.append_entries([dup_entry, dup_entry2])
    expected_matches = [ind, structure_wrangler.num_structures - 2]
    print(structure_wrangler.get_matching_corr_duplicate_indices(), expected_matches)
    assert any(
        expected_matches == sorted(matches)
        for matches in structure_wrangler.get_matching_corr_duplicate_indices()
    )


def test_get_constant_features(structure_wrangler, rng):
    ind = rng.integers(1, structure_wrangler.num_features)
    for entry in structure_wrangler.entries:
        entry.data["correlations"][ind] = 3.0  # make constant
    assert ind in structure_wrangler.get_constant_features()
    assert 0 not in structure_wrangler.get_constant_features()


def test_msonable(structure_wrangler):
    structure_wrangler.metadata["key"] = 4
    d = structure_wrangler.as_dict()
    sw = StructureWrangler.from_dict(d)
    npt.assert_array_equal(
        sw.get_property_vector("energy"),
        structure_wrangler.get_property_vector("energy"),
    )
    npt.assert_array_equal(
        sw.get_weights("random"), structure_wrangler.get_weights("random")
    )
    assert all(
        [s1 == s2 for s1, s2 in zip(sw.structures, structure_wrangler.structures)]
    )
    assert all(
        [
            s1 == s2
            for s1, s2 in zip(
                sw.refined_structures, structure_wrangler.refined_structures
            )
        ]
    )
    npt.assert_array_equal(sw.feature_matrix, structure_wrangler.feature_matrix)
    assert sw.metadata == structure_wrangler.metadata
    assert all(
        i1.data["site_mapping"] == i1.data["site_mapping"]
        for i1, i2 in zip(structure_wrangler._entries, sw._entries)
    )
    assert all(
        np.array_equal(m1, m2)
        for m1, m2 in zip(structure_wrangler.supercell_matrices, sw.supercell_matrices)
    )
    assert_msonable(structure_wrangler)



"""Implementation of the Cluster class.

Represents a group of sites of a given lattice. These are the building blocks
for a cluster basis of functions over configurational space.
"""

__author__ = "Luis Barroso-Luque, William Davidson Richard"

import json
import os
from fnmatch import fnmatch
from functools import cached_property
from io import StringIO

import numpy as np
from monty.io import zopen
from monty.json import MSONable
from pymatgen.core import Lattice, Site
from pymatgen.core.structure import Composition, SiteCollection
from pymatgen.util.coord import is_coord_subset
from ruamel.yaml import YAML

from smol.cofe.space.constants import SITE_TOL
from smol.cofe.space.domain import Vacancy, get_site_spaces


class Cluster(SiteCollection, MSONable):
    """An undecorated (no occupancies) cluster.

    Represented simply by a list of sites, its centroid, and the underlying
    lattice.

    You probably never need to instantiate this class directly. Look at
    ClusterSubspace to create orbits and clusters necessary for a CE.

    Attributes:
        frac_coords (ndarray): fractional coordinates of each site.
        lattice (Lattice): Underlying lattice of cluster.
        centroid (float): Geometric centroid of included sites.
        id (int): ID of cluster.
            Used to identify the Cluster in a given ClusterSubspace.
    """

    def __init__(self, site_spaces, frac_coords, lattice):
        """Initialize Cluster.

        Args:
            site_spaces (list of SiteSpace):
                list of site spaces for the cluster
            frac_coords (Sequence):
                Sequence of frac coords for the site spaces
            lattice (Lattice):
                pymatgen Lattice object
        """
        frac_coords = np.array(frac_coords)
        centroid = np.average(frac_coords, axis=0)
        shift = np.floor(centroid)
        self._centroid = centroid - shift
        self._frac_coords = frac_coords - shift
        self._sites = tuple(
            Site(site_space, coords)
            for site_space, coords in zip(
                site_spaces, lattice.get_cartesian_coords(frac_coords)
            )
        )
        self._lattice = lattice
        self.id = None

    @property
    def centroid(self):
        """Return the centroid of cluster."""
        return self._centroid

    @property
    def frac_coords(self):
        """Return the fractional coordinates of cluster w.r.t the underlying lattice."""
        return self._frac_coords

    @cached_property
    def diameter(self):
        """Get maximum distance between any 2 sites in the cluster."""
        coords = self.lattice.get_cartesian_coords(self.frac_coords)
        all_d2 = np.sum((coords[None, :, :] - coords[:, None, :]) ** 2, axis=-1)
        return np.max(all_d2) ** 0.5

    @property
    def radius(self):
        """Get half the maximum distance between any 2 sites in the cluster."""
        return self.diameter / 2.0

    @property
    def lattice(self):
        """Return the underlying lattice."""
        return self._lattice

    @property
    def sites(self):
        """Return the list of sites."""
        return self._sites

    def get_distance(self, i: int, j: int) -> float:
        """Return distance between sites at index i and j.

        Args:
            i: Index of first site
            j: Index of second site
        Returns:
            Distance between sites at index i and index j.
        """
        return self[i].distance(self[j])

    def assign_ids(self, cluster_id):
        """Recursively assign IDs to clusters after initialization."""
        self.id = cluster_id
        return cluster_id + 1

    def to(self, fmt: str = None, filename: str = None):
        """
        Output the cluster  to a file or string.

        this is basically a watered down version of pymatgen.Molecule.to

        Args:
            fmt (str): Format to output to. Defaults to JSON unless filename
                is provided. If fmt is specifies, it overrides whatever the
                filename is. Options include "yaml" and "json" only.
                 Non-case sensitive.
            filename (str): If provided, output will be written to a file. If
                fmt is not specified, the format is determined from the
                filename. Defaults is None, i.e. string output.
        Returns:
            (str) if filename is None. None otherwise.
        """
        fmt = "" if fmt is None else fmt.lower()
        fname = os.path.basename(filename or "")
        if fmt == "json" or fnmatch(fname, "*.json*") or fnmatch(fname, "*.mson*"):
            if filename:
                with zopen(filename, "wt", encoding="utf8") as f:
                    return json.dump(self.as_dict(), f)
            else:
                return json.dumps(self.as_dict())

        if fmt == "yaml" or fnmatch(fname, "*.yaml*"):
            yaml = YAML()
            if filename:
                with zopen(fname, "wt", encoding="utf8") as f:
                    return yaml.dump(self.as_dict(), f)
            else:
                sio = StringIO()
                yaml.dump(self.as_dict(), sio)
                return sio.getvalue()

        raise ValueError(f"Invalid format: `{str(fmt)}`")

    @classmethod
    def from_str(cls, input_string: str, fmt):
        """
        Read a cluster from a string.

        Args:
            input_string (str): String to parse.
            fmt (str): Format to output to. Defaults to JSON unless filename
                is provided. If fmt is specifies, it overrides whatever the
                filename is. Options include "yaml", "json". Non-case sensitive.
        Returns:
            Cluster
        """
        if fmt == "json":
            d = json.loads(input_string)
            return cls.from_dict(d)
        if fmt == "yaml":
            yaml = YAML()
            d = yaml.load(input_string)
            return cls.from_dict(d)

        raise ValueError(f"Invalid format: `{str(fmt)}`")

    @classmethod
    def from_file(cls, filename: str):
        """
        Read a cluster from a file. Supported formats are json and yaml only.

        Args:
            filename (str): The filename to read from.
        Returns:
            Cluster
        """
        filename = str(filename)

        with zopen(filename) as f:
            contents = f.read()
        fname = filename.lower()

        if fnmatch(fname, "*.json*") or fnmatch(fname, "*.mson*"):
            return cls.from_str(contents, fmt="json")
        if fnmatch(fname, "*.yaml*"):
            return cls.from_str(contents, fmt="yaml")

        raise ValueError("Cannot determine file type.")

    @classmethod
    def from_sites(cls, sites, lattice):
        """Create a cluster from a list of sites and lattice object."""
        frac_coords = [lattice.get_fractional_coords(site.coords) for site in sites]
        site_spaces = get_site_spaces(sites)
        return cls(site_spaces, frac_coords, lattice)

    def __eq__(self, other):
        """Check equivalency of clusters considering symmetry."""
        if self.frac_coords.shape != other.frac_coords.shape:
            return False
        othersites = other.frac_coords + np.round(self.centroid - other.centroid)
        return is_coord_subset(self.frac_coords, othersites, atol=SITE_TOL)

    def __str__(self):
        """Pretty print a cluster."""
        centroid_str = " ".join(
            [
                f"{j:0.6f}".rjust(12)
                for j in self.lattice.get_cartesian_coords(self.centroid)
            ]
        )
        centroid_str += "  -> " + " ".join(
            [f"{j:0.6f}".rjust(12) for j in self.centroid]
        )
        outs = [
            f"Diameter : {self.diameter:0.4f}",
            f"  Charge : {self.charge}",
            f"Centroid : {centroid_str}",
            f"Sites ({len(self)})",
        ]
        site_outs = []
        for i, site in enumerate(self):
            site_outs.append(
                " ".join(
                    [
                        str(i),
                        site.species_string,
                        " ".join([f"{j:0.6f}".rjust(12) for j in site.coords]),
                        "  ->",
                        " ".join(
                            [
                                f"{j:0.6f}".rjust(12)
                                for j in self.lattice.get_fractional_coords(site.coords)
                            ]
                        ),
                    ]
                )
            )
        width = len(site_outs[0])
        outs.append(width * "-")
        outs += site_outs
        return "\n".join(outs)

    def __repr__(self):
        """Get cluster summary."""
        centroid_str = "({:.4f}, {:.4f}, {:.4f})".format(
            *self.lattice.get_cartesian_coords(self.centroid)
        ) + " -> [{:.4f}, {:.4f}, {:.4f}]".format(*self.centroid)
        outs = [
            "Cluster",
            f"No. sites: {len(self._sites)}   Diameter: {self.diameter:0.4f}",
            f"Centroid: {centroid_str}",
        ]
        for s in self:
            outs.append(
                s.__repr__()
                + " -> [{:.4f}, {:.4f}, {:.4f}]".format(
                    *self.lattice.get_fractional_coords(s.coords)
                )
            )
        return "\n".join(outs)

    @classmethod
    def from_dict(cls, d):
        """Create a Cluster from serialized dict."""
        sites = [Site.from_dict(item) for item in d["sites"]]
        # Force vacancies back to vacancies
        for symbols, site in zip(d["vacancy_symbols"], sites):
            site.species = Composition(
                {
                    spec
                    if spec.symbol not in symbols
                    else Vacancy(
                        spec.symbol, spec.oxidation_state, spec.properties
                    ): val
                    for spec, val in site.species.items()
                    if spec.symbol not in symbols
                }
            )

        cluster = Cluster.from_sites(sites, Lattice.from_dict(d["lattice"]))
        #  force set this to get exact object
        cluster._frac_coords = np.array(d["frac_coords"])
        return cluster

    def as_dict(self):
        """Get json-serialization dict representation.

        Returns:
            MSONable dict
        """
        cluster_d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
            "lattice": self.lattice.as_dict(),
            "frac_coords": self.frac_coords.tolist(),
            "sites": [site.as_dict() for site in self.sites],
            "vacancy_symbols": [
                [spec.symbol for spec in site.species if isinstance(spec, Vacancy)]
                for site in self.sites
            ],
        }
        return cluster_d


# functions to render cluster using crystal-toolkit
try:
    from collections import defaultdict

    from crystal_toolkit.core.legend import Legend
    from crystal_toolkit.core.scene import Scene

    def get_cluster_scene(
        self,
        origin=None,
        legend=None,
    ) -> Scene:
        """
        Create CTK objects for the lattice and sties.

        Args:
            self:  Structure object
            origin: coordinate of the origin
            legend: Legend for the sites

        Returns:
            CTK scene object to be rendered
        """
        origin = origin or list(-self.lattice.get_cartesian_coords([0.5, 0.5, 0.5]))
        legend = legend or Legend(self)
        primitives = defaultdict(list)

        for site in self:
            site_scene = site.get_scene(
                origin=origin,
                legend=legend,
            )
            for scene in site_scene.contents:
                primitives[scene.name] += scene.contents

        primitives["unit_cell"].append(self.lattice.get_scene())

        return Scene(
            name="Cluster",
            origin=origin,
            contents=[
                Scene(name=k, contents=v, origin=origin) for k, v in primitives.items()
            ],
        )

    Cluster.get_scene = get_cluster_scene

except ImportError:
    pass



"""Tools for training structure selection."""

from warnings import warn

import numpy as np
from scipy.linalg import lu, orth
from scipy.stats import multinomial, multivariate_normal

__author__ = "Luis Barroso-Luque"


def full_row_rank_select(feature_matrix, tol=1e-15, nrows=None):
    """Choose a (maximally) full rank subset of rows in a feature matrix.

    This method is for underdetermined systems, i.e., where
    columns (i.e. features) > rows (i.e. structures)

    Args:
        feature_matrix (ndarray):
            feature matrix to select rows/structures from.
        tol (float): optional
            tolerance to use to determine the pivots in upper triangular matrix
            of the LU decomposition.
        nrows (int): optional
            number of rows to include. If None, will include the maximum
            possible number of rows.
    Returns:
        list of int: list with indices of rows that form a full rank system.
    """
    nrows = nrows if nrows is not None else feature_matrix.shape[0]
    _, _, u_mat = lu(feature_matrix.T)  # pylint: disable=unbalanced-tuple-unpacking
    sample_ids = np.unique((abs(u_mat) > tol).argmax(axis=1))[:nrows]
    if len(sample_ids) > np.linalg.matrix_rank(feature_matrix[:nrows]):
        warn(
            "More samples than the matrix rank where selected.\n"
            "The selection will not actually be full rank.\n"
            "Decrease tolerance to ensure full rank indices are selected.\n"
        )
    return np.sort(sample_ids)


def gaussian_select(feature_matrix, num_samples, orthogonalize=False):
    """V. Vanilla structure selection to increase incoherence.

    Sequentially picks samples with feature vectors that most closely aligns
    with a sampled random gaussian vector on the unit sphere.

    This works much better when the number of rows in the feature matrix
    is much larger than the number of samples requested.

    Args:
        feature_matrix (ndarray):
            feature matrix to select rows/structures from.
        num_samples (int):
            number of samples/rows/structures to select.
        orthogonalize (bool):
            if true, will orthogonalize the generated random vectors
    Returns:
        list of int: list with indices of rows that align with Gaussian samples
    """
    matrix = feature_matrix.copy()
    rows, cols = matrix.shape
    matrix /= np.linalg.norm(matrix, axis=1).reshape(rows, 1)
    gauss = multivariate_normal(mean=np.zeros(cols)).rvs(size=num_samples)
    if orthogonalize:
        gauss = orth(gauss.T).T  # orthogonalize rows of gauss

    gauss /= np.linalg.norm(gauss, axis=1).reshape(num_samples, 1)
    sample_ids, compliment = [], list(range(rows))
    for vec in gauss:
        sample_ids.append(compliment[np.argmax(matrix[compliment] @ vec)])
        compliment.remove(sample_ids[-1])
    return np.sort(sample_ids)


def composition_select(
    composition_vector, composition, cell_sizes, num_samples, rng=None
):
    """Structure selection based on composition multinomial probability.

    Note: this function is needs quite a bit of tweaking to get nice samples.

    Args:
        composition_vector (ndarray):
            N (samples) by n (components) with the composition of the samples
            to select from.
        composition (ndarray):
            array for the center composition to sample around.
        cell_sizes (int or Sequence):
            mumber of unit cells or size of supercells used to set the
            number of variables in the multinomial distribution.
        num_samples (int):
            number of samples to return. Note that if the number is too high
            compared to the total number of samples (or coverage of the space),
            it may take very long to return if at all, and the samples will not
            be very representative of the multinomial distributions.
        rng (np.Generator): optional
            numpy seed, Generator, SeedSequence, etc

    Returns:
        list of int: list with indices of the composition vector corresponding
                     to selected samples.
    """
    rng = np.random.default_rng(rng)
    unique_compositions = np.unique(composition_vector, axis=1)
    # Sample structures biased by composition
    if isinstance(cell_sizes, int):
        cell_sizes = np.array(
            len(composition_vector)
            * [
                cell_sizes,
            ]
        )
    else:
        cell_sizes = np.array(cell_sizes)

    dists = {size: multinomial(size, composition) for size in np.unique(cell_sizes)}
    max_probs = {
        size: dist.pmf(size * unique_compositions).max() for size, dist in dists.items()
    }

    sample_ids = [
        i
        for i, (size, comp) in enumerate(zip(cell_sizes, composition_vector))
        if dists[size].pmf(size * comp) >= max_probs[size] * rng.random()
    ]
    while len(sample_ids) <= num_samples:
        sample_ids += [
            i
            for i, (size, comp) in enumerate(zip(cell_sizes, composition_vector))
            if dists[size].pmf(size * comp) >= max_probs[size] * rng.random()
            and i not in sample_ids
        ]
    return np.sort(sample_ids[:num_samples])



from collections import OrderedDict

import numpy as np
import numpy.testing as npt
import pytest
from numpy.polynomial.chebyshev import chebval
from numpy.polynomial.legendre import legval
from pymatgen.core import Composition

from smol.cofe.space import basis, domain
from smol.utils import get_subclasses
from tests.utils import assert_msonable

basis_iterators = list(get_subclasses(basis.BasisIterator).values())


@pytest.fixture(params=[True, False], scope="module")
def site_space(expansion_structure, request):
    spaces = domain.get_site_spaces(expansion_structure, include_measure=request.param)
    return spaces[0]


@pytest.fixture(params=basis_iterators)
def standard_basis(site_space, request):
    species = tuple(site_space.keys())
    return basis.StandardBasis(site_space, request.param(species))


@pytest.mark.parametrize("basis_iterator_cls", basis_iterators)
def test_basis_measure_raise(basis_iterator_cls):
    comp = Composition((("A", 0.25), ("B", 0.25), ("C", 0.8), ("D", 0.25)))
    with pytest.raises(ValueError):
        space = domain.SiteSpace(comp)
        species = tuple(space.keys())
        basis.StandardBasis(space, basis_iterator_cls(species))


@pytest.mark.parametrize("basis_iterator_cls", basis_iterators)
def test_constructor(basis_iterator_cls):
    species = OrderedDict({"A": 1, "B": 2, "C": 1})
    basis_iter = basis_iterator_cls(tuple(species.keys()))
    with pytest.warns(RuntimeWarning):
        basis.StandardBasis(species, basis_iter)

    species = OrderedDict({"A": 0.1, "B": 0.1, "C": 0.1})
    basis_iter = basis.SinusoidIterator(tuple(species.keys()))

    with pytest.warns(RuntimeWarning):
        basis.StandardBasis(species, basis_iter)
    basis_iter = basis_iterator_cls(tuple(species.keys())[:-1])

    with pytest.raises(ValueError):
        basis.StandardBasis(species, basis_iter)


@pytest.mark.parametrize("basis_iterator_cls", basis_iterators)
def test_msonable(site_space, basis_iterator_cls):
    species = tuple(site_space.keys())
    b = basis.StandardBasis(site_space, basis_iterator_cls(species))
    assert_msonable(b)
    b1 = basis.StandardBasis.from_dict(b.as_dict())
    assert b.is_orthogonal == b1.is_orthogonal
    assert b.is_orthonormal == b1.is_orthonormal
    npt.assert_array_equal(b.function_array, b1.function_array)
    npt.assert_array_equal(b.orthonormalization_array, b1.orthonormalization_array)

    b.orthonormalize()
    b1 = basis.StandardBasis.from_dict(b.as_dict())
    assert b1.is_orthogonal
    assert b1.is_orthonormal
    npt.assert_array_equal(b.function_array, b1.function_array)
    npt.assert_array_equal(b.orthonormalization_array, b1.orthonormalization_array)


def test_standard_basis(standard_basis):
    site_space = standard_basis.site_space
    measure = list(standard_basis.measure_vector)
    assert measure == list(site_space.values())

    original = standard_basis._f_array.copy()
    standard_basis.orthonormalize()
    assert standard_basis.is_orthonormal
    assert np.allclose(standard_basis._r_array @ standard_basis._f_array, original)


def test_rotate(standard_basis, rng):
    f_array = standard_basis._f_array.copy()
    theta = np.pi / rng.integers(2, 10)

    # catch and test non-uniform measures
    if not np.allclose(standard_basis.measure_vector[0], standard_basis.measure_vector):
        with pytest.warns(UserWarning):
            standard_basis.rotate(theta)
        return

    standard_basis.rotate(theta)
    if standard_basis.is_orthogonal:
        assert standard_basis.is_orthogonal
        # if orthogonal all inner products should match!
        npt.assert_array_almost_equal(
            f_array @ f_array.T, standard_basis._f_array @ standard_basis._f_array.T
        )
    else:
        # if not orthogonal only vector norms should mathch!
        npt.assert_array_almost_equal(
            np.diag(f_array @ f_array.T),
            np.diag(standard_basis._f_array @ standard_basis._f_array.T),
        )

    npt.assert_array_almost_equal(
        standard_basis._f_array[0], np.ones(len(standard_basis.site_space))
    )
    if len(standard_basis.site_space) == 2:  # rotation is * -1
        npt.assert_array_almost_equal(standard_basis._f_array[1], -1 * f_array[1])
    else:
        npt.assert_almost_equal(
            np.arccos(
                np.dot(standard_basis._f_array[1], f_array[1])
                / (
                    np.linalg.norm(standard_basis._f_array[1])
                    * np.linalg.norm(f_array[1])
                )
            ),
            theta,
        )
    standard_basis.rotate(-theta)
    npt.assert_array_almost_equal(f_array, standard_basis._f_array)
    standard_basis.orthonormalize()
    assert standard_basis.is_orthonormal
    standard_basis.rotate(theta)
    assert standard_basis.is_orthonormal

    if len(standard_basis.site_space) > 2:
        with pytest.raises(ValueError):
            standard_basis.rotate(theta, 0, 0)

        with pytest.raises(ValueError):
            standard_basis.rotate(theta, len(standard_basis.site_space))
        with pytest.raises(ValueError):
            standard_basis.rotate(theta, 0, len(standard_basis.site_space))

    comp = Composition((("A", 0.2), ("B", 0.2), ("C", 0.3), ("D", 0.3)))
    b = basis.basis_factory("sinusoid", domain.SiteSpace(comp))
    with pytest.warns(UserWarning):
        b.rotate(theta)


# basis specific tests
def test_indicator_basis(site_space):
    species = tuple(site_space.keys())
    b = basis.StandardBasis(site_space, basis.IndicatorIterator(species))
    assert not b.is_orthogonal

    # test evaluation of basis functions
    n = len(site_space) - 1
    for i in range(n):
        assert b.function_array[i, i] == 1

    original = b._f_array.copy()
    b.orthonormalize()
    assert b.is_orthonormal
    npt.assert_allclose(b._r_array @ b._f_array, original, atol=5e-16)


def test_sinusoid_basis(site_space):
    species = tuple(site_space.keys())
    b = basis.StandardBasis(site_space, basis.SinusoidIterator(species))

    # ortho only holds for uniform measure
    if np.allclose(b.measure_vector[0], b.measure_vector):
        assert b.is_orthogonal

        if len(site_space) == 2:
            assert b.is_orthonormal
        else:
            assert not b.is_orthonormal

    # test evaluation of basis functions
    m = len(site_space)
    for n in range(1, len(site_space)):
        a = -(-n // 2)
        f = (
            lambda s: -np.sin(2 * np.pi * a * s / m)
            if n % 2 == 0
            else -np.cos(2 * np.pi * a * s / m)
        )
        for i, _ in enumerate(site_space):
            npt.assert_almost_equal(b.function_array[n - 1, i], f(i))

    original = b._f_array.copy()
    b.orthonormalize()
    assert b.is_orthonormal
    npt.assert_allclose(b._r_array @ b._f_array, original, atol=1e-15)


def test_chebyshev_basis(site_space):
    species = tuple(site_space.keys())
    b = basis.StandardBasis(site_space, basis.ChebyshevIterator(species))
    if len(site_space) == 2:
        assert b.is_orthogonal
    else:
        assert not b.is_orthogonal

    # test evaluation of basis functions
    m, coeffs = len(site_space), [1]
    fun_range = np.linspace(-1, 1, m)
    for n in range(m - 1):
        coeffs.append(0)
        for sp, x in enumerate(fun_range):
            assert b.function_array[n, sp] == chebval(x, c=coeffs[::-1])
    b.orthonormalize()
    assert b.is_orthonormal


def test_legendre_basis(site_space):
    species = tuple(site_space.keys())
    b = basis.StandardBasis(site_space, basis.LegendreIterator(species))
    if len(site_space) == 2:
        assert b.is_orthogonal
    else:
        assert not b.is_orthogonal

    # test evaluation of basis functions
    m, coeffs = len(site_space), [1]
    fun_range = np.linspace(-1, 1, m)
    for n in range(m - 1):
        coeffs.append(0)
        for sp, x in enumerate(fun_range):
            assert b.function_array[n, sp] == legval(x, c=coeffs[::-1])

    b.orthonormalize()
    assert b.is_orthonormal



"""Functions to filter data in a StructureWrangler and calculate fitting weights.

For example, weights can be calculated by energy above hull or energy by composition.
"""

from collections import defaultdict

import numpy as np
from pymatgen.analysis.phase_diagram import PDEntry, PhaseDiagram
from pymatgen.core import Composition

from smol.cofe.extern import EwaldTerm

__author__ = "Luis Barroso-Luque, William Davidson Richard"

from smol.constants import kB


def unique_corr_vector_indices(
    wrangler, property_key, filter_by="min", cutoffs=None, return_compliment=False
):
    """Return set of indices of structures with unique correlation vectors.

    Keep structures with the min or max value of the given property.
    Note correlation vectors exclude external terms such that even if the
    external term is different for the structure but all correlations are
    the same, the structures will be considered duplicates.

    Args:
        wrangler (StructureWrangler):
            a StructureWrangler containing data to be filtered.
        property_key (str):
            name of property to consider when returning structure indices
            for the feature matrix with duplicate corr vectors removed.
        filter_by (str)
            the criteria for the property value to keep. Options are min
            or max.
        cutoffs (dict): optional
            Dictionary with cluster diameter cutoffs for correlation
            functions to consider in correlation vectors.
        return_compliment (bool): optional
            if True, will return the complement of the unique indices
    Returns:
        indices, complement
    """
    if filter_by not in ("max", "min"):
        raise ValueError(f"Filtering by {filter_by} is not an option.")

    choose_val = np.argmin if filter_by == "min" else np.argmax

    property_vector = wrangler.get_property_vector(property_key)
    dupe_inds = wrangler.get_duplicate_corr_indices(cutoffs=cutoffs)
    indices = {inds[choose_val(property_vector[inds])] for inds in dupe_inds}
    duplicates = {i for inds in dupe_inds for i in inds} - indices
    indices = list(set(range(wrangler.num_structures)) - duplicates)

    if return_compliment:
        return indices, list(duplicates)
    return indices


def max_ewald_energy_indices(wrangler, max_relative_energy, return_compliment=False):
    """Return set of indices of structures by electrostatic interaction energy.

    Filter the input structures to remove those with low electrostatic
    energies (no large charge separation in cell). This energy is
    referenced to the lowest value at that composition. Note that this is
    before the division by the relative dielectric constant and is per
    primitive cell in the cluster expansion -- 1.5 eV/atom seems to be a
    reasonable value for dielectric constants around 10.

    Args:
        wrangler (StructureWrangler):
            a StructureWrangler containing data to be filtered.
        max_relative_energy (float):
            Ewald threshold. The maximum Ewald energy relative to the
            minimum energy at a structure's composition (normalized per prim).
        return_compliment (bool): optional
            If True will return the compliment of the unique indices.
    Returns:
        indices, compliment
    """
    ewald_energy = None
    for term in wrangler.cluster_subspace.external_terms:
        if isinstance(term, EwaldTerm):
            ewald_energy = [
                entry.data["correlations"][-1] for entry in wrangler.entries
            ]
    if ewald_energy is None:
        ewald_energy = []
        for entry in wrangler.entries:
            struct = entry.structure
            matrix = entry.data["supercell_matrix"]
            mapping = entry.data["site_mapping"]
            occu = wrangler.cluster_subspace.occupancy_from_structure(
                struct, encode=True, scmatrix=matrix, site_mapping=mapping
            )

            supercell = wrangler.cluster_subspace.structure.copy()
            supercell.make_supercell(matrix)
            size = wrangler.cluster_subspace.num_prims_from_matrix(matrix)
            term = EwaldTerm().value_from_occupancy(occu, supercell) / size
            ewald_energy.append(term)

    min_energy_by_comp = defaultdict(lambda: np.inf)
    for energy, entry in zip(ewald_energy, wrangler.entries):
        comp = entry.structure.composition.reduced_composition
        if energy < min_energy_by_comp[comp]:
            min_energy_by_comp[comp] = energy

    indices, compliment = [], []
    for i, (energy, entry) in enumerate(zip(ewald_energy, wrangler.entries)):
        min_energy = min_energy_by_comp[entry.structure.composition.reduced_composition]
        rel_energy = energy - min_energy
        if rel_energy < max_relative_energy:
            indices.append(i)
        else:
            compliment.append(i)

    if return_compliment:
        return indices, compliment
    return indices


def weights_energy_above_composition(structures, energies, temperature=2000):
    """Compute weights for energy above the minimum reduced composition energy.

    Args:
        structures (list):
            list of pymatgen Structures
        energies (ndarray):
            energies of corresponding Structures
        temperature (float):
            temperature to use in Boltzmann weight

    Returns: weights for each structure.
        array
    """
    e_above_comp = _energies_above_composition(structures, energies)
    return np.exp(-e_above_comp / (kB * temperature))


def weights_energy_above_hull(structures, energies, cs_structure, temperature=2000):
    """Compute weights for structure energy above the hull of given structures.

    Args:
        structures (list):
            list of pymatgen Structures
        energies (ndarray):
            energies of corresponding Structures.
        cs_structure (Structure):
            The pymatgen Structure used to define the ClusterSubspace
        temperature (float):
            temperature to use in Boltzmann weight.

    Returns: weights for each structure.
        array
    """
    e_above_hull = _energies_above_hull(structures, energies, cs_structure)
    return np.exp(-e_above_hull / (kB * temperature))


def _energies_above_hull(structures, energies, ce_structure):
    """Compute energies above hull.

    The hull is constructed from the phase diagram of the given structures.
    """
    phase_diagram = _pd(structures, energies, ce_structure)
    e_above_hull = []
    for structure, energy in zip(structures, energies):
        entry = PDEntry(structure.composition.element_composition, energy)
        e_above_hull.append(phase_diagram.get_e_above_hull(entry))
    return np.array(e_above_hull)


def _pd(structures, energies, cs_structure):
    """Generate a phase diagram with the structures and energies."""
    entries = []

    for structure, energy in zip(structures, energies):
        entries.append(PDEntry(structure.composition.element_composition, energy))

    max_e = max(entries, key=lambda e: e.energy_per_atom).energy_per_atom
    max_e += 1000
    for element in cs_structure.composition.keys():
        entry = PDEntry(Composition({element: 1}).element_composition, max_e)
        entries.append(entry)

    return PhaseDiagram(entries)


def _energies_above_composition(structures, energies):
    """Compute structure energies above reduced composition."""
    min_e = defaultdict(lambda: np.inf)
    for structure, energy in zip(structures, energies):
        comp = structure.composition.reduced_composition
        if energy / len(structure) < min_e[comp]:
            min_e[comp] = energy / len(structure)
    e_above = []
    for structure, energy in zip(structures, energies):
        comp = structure.composition.reduced_composition
        e_above.append(energy / len(structure) - min_e[comp])
    return np.array(e_above)



import json
import os

import numpy as np
from pymatgen.core.structure import Structure

DATA_DIR = os.path.join(os.path.dirname(__file__), "..", "data")


# LNO example data
with open(os.path.join(DATA_DIR, "lno_prim.json")) as f:
    lno_prim = Structure.from_dict(json.load(f))

with open(os.path.join(DATA_DIR, "lno_fitting_data.json")) as f:
    lno_data = [(Structure.from_dict(x["s"]), x["toten"]) for x in json.load(f)]

with open(os.path.join(DATA_DIR, "lno_gc_run.json")) as f:
    lno_gc_run_10000 = json.load(f)

# Pyabinitio benchmark datasets
with open(os.path.join(DATA_DIR, "pyabinitio-LiMn3OF.json")) as f:
    pyabinitio_LiMn3OF_dataset = json.load(f)
    pyabinitio_LiMn3OF_dataset["prim"] = Structure.from_dict(
        pyabinitio_LiMn3OF_dataset["prim"]
    )
    pyabinitio_LiMn3OF_dataset["ecis"] = np.array(pyabinitio_LiMn3OF_dataset["ecis"])
    pyabinitio_LiMn3OF_dataset["feature_matrix"] = np.array(
        pyabinitio_LiMn3OF_dataset["feature_matrix"]
    )
    for item in pyabinitio_LiMn3OF_dataset["data"]:
        item["structure"] = Structure.from_dict(item["structure"])
        item["scmatrix"] = np.array(item["scmatrix"])

with open(os.path.join(DATA_DIR, "pyabinitio-LiNi2Mn4OF.json")) as f:
    pyabinitio_LiNi2Mn4OF_dataset = json.load(f)
    pyabinitio_LiNi2Mn4OF_dataset["prim"] = Structure.from_dict(
        pyabinitio_LiNi2Mn4OF_dataset["prim"]
    )
    pyabinitio_LiNi2Mn4OF_dataset["ecis"] = np.array(
        pyabinitio_LiNi2Mn4OF_dataset["ecis"]
    )
    pyabinitio_LiNi2Mn4OF_dataset["feature_matrix"] = np.array(
        pyabinitio_LiNi2Mn4OF_dataset["feature_matrix"]
    )
    for item in pyabinitio_LiNi2Mn4OF_dataset["data"]:
        item["structure"] = Structure.from_dict(item["structure"])
        item["scmatrix"] = np.array(item["scmatrix"])

# icet AuPd benchmark data
with open(os.path.join(DATA_DIR, "benchmark_data/icet_aupd", "aupd_prim.json")) as f:
    aupt_prim = Structure.from_dict(json.load(f))

with open(
    os.path.join(DATA_DIR, "benchmark_data/icet_aupd", "aupd_linreg_eci.json")
) as f:
    icet_eci = np.array(json.load(f))

with open(
    os.path.join(DATA_DIR, "benchmark_data/icet_aupd", "aupd_predictions.json")
) as f:
    icet_predictions = np.array(json.load(f))

with open(os.path.join(DATA_DIR, "benchmark_data/icet_aupd", "fit_structs.json")) as f:
    icet_fit_structs = json.load(f)
    for item in icet_fit_structs:
        item["structure"] = Structure.from_dict(item["structure"])
        item["scmatrix"] = np.array(item["scmatrix"])

with open(os.path.join(DATA_DIR, "benchmark_data/icet_aupd", "test_structs.json")) as f:
    icet_test_structs = json.load(f)
    for item in icet_test_structs:
        item["structure"] = Structure.from_dict(item["structure"])
        item["scmatrix"] = np.array(item["scmatrix"])

with open(os.path.join(DATA_DIR, "benchmark_data/icet_aupd", "sgc_run.json")) as f:
    icet_sgc_run_10000its = json.load(f)
